/**
  @file
  Sceptre main.
 */

/*! \mainpage Sceptre Preloaded Test Application
 *
 * \section version_sec Version
 * Clemens Valens, January 31, 2010
 *
 * \section intro_sec Introduction
 *
 * This is a test program that will try to create some files
 * filled with data on a preformatted SD card that shows that
 * everything is working OK.
 *
 * Note: Some SD cards need a second board reset before they work. (Why?)
 *
 * Some information about the system is written to "sys_info.txt".
 *
 * Some information about the SD card itself is written to "sd_info.txt".
 *
 * The board will also output data on UART0 (115200n81). Instructions
 * and initialization progress messages are transmitted. 
 *
 * After initialization the system will log at 10 Hz during 30 seconds
 * the accelerometer output to the file "a_log.txt" and the thermometer 
 * output to "t_log.txt". These log files also contain a second counter
 * that shows that the real time clock (RTC) is running. A progress indicator
 * is transmitted on UART0.
 *
 * When the data logging terminates
 * the system will transmit a message every second with time,
 * temperature and accelerometer data. It will also launch a USB mass storage device.
 *
 * The SD card can be accessed as a USB mass storage device as soon as
 * the USB connect LED is lighted. On-the-fly USB port switching is possible
 * by first removing JP5, then JP4, then repositioning JP4 and finish by
 * repositioning JP5. This works on Windows XP Pro. Switching back to USB serial port
 * will very probably reset the board.
 * 
 * An LED connected between port 1.30 (anode, processor pin 52, Sceptre K7 pin 1)
 * and GND (cathode) will blink at a rate of 1 Hz under tilmer interrupt control.
 *
 * Since the basic card does not have a Bluetooth module mounted,
 * its test is not activated by default. A test is however available 
 * To activate Bluetooth testing provide a file named "test_bt.pls" on the SD card.
 * The file should contain the friendly name for the Bluetooth device (max 16 characters).
 *
 * The Bluetooth test will put the BT module in slave connection mode (link LED will blink).
 * It should be possible to connect to the board using the PIN or ID
 * "1234". A connection changes the behaviour of the BT LEDs (link LED will stay on),
 * a disconnect message is transmitted on UART0 when the connection is lost.
 *
 * After data logging the system will transmit data received on UART0 to the Bluetooth peer,
 * data received on the Bluetooth module is echoed on UART0, interlaced with the
 * time-temperature-accelerometer message. Data reception is checked at 1 Hz.
 *
 * The BT module runs at 19200n81.
 */


#include "sceptre.h"

#include <time.h>  // For CLOCKS_PER_SEC
#include <string.h>
#include <stdio.h>
#include "..\core\filesystem\sdcard_driver.h"
#include "..\core\filesystem\sdcard.h"
#include "..\core\thermometer\ds18b20.h"


const char *p_hello = "Sceptre rulez!\n";
const char *p_version = "V2010.01.31\n\n";
const char *p_name1 = "a_log.txt";
const char *p_name2 = "t_log.txt";
const char *p_bluetooth_test = "test_bt.pls"; // Use a "please" file to activate bluetooth test.
rtc_time_t rtc;  // A global RTC structure.
char bt_name[16+1];


void main_send_string(const char *p_str);
void main_10ms_timer(void);
FIL *main_file_open(FIL *p_file, const char *p_name, uint8_t flags);
bool_t main_file_exist(const char *p_name);
void main_log_data(const char *p_name_1, const char *p_name_2);


#define LED_PIN  30
#define LED_ON  IOSET1 = (1<<LED_PIN)
#define LED_OFF  IOCLR1 = (1<<LED_PIN)
void led_init(void);
void led_tick(void);

void led_init(void)
{
  IODIR1 |= (1<<LED_PIN);  // P1.30 is output.
  LED_OFF;  // LED off.
}

void led_tick(void)
{
  static int ms = 0;

  ms += 10;
  if (ms>=500)
  {
    ms = 0;
    if ((IOPIN1&(1<<LED_PIN))==0) LED_ON;
    else LED_OFF;
  }
}


void main_send_string(const char *p_str)
{
  uart0_send((const uint8_t *)p_str,strlen(p_str),true);
}


void main_10ms_timer(void)
{
  disk_timerproc();
  led_tick();
}


FIL *main_file_open(FIL *p_file, const char *p_name, uint8_t flags)
{
  if (f_open(p_file,p_name,flags)!=FR_OK)
  {
    /// \todo Figure out why it crashes here.
    printf("Could not open file \"%s\"\n",p_name);
    return NULL;
  }
  return p_file;
}


/// \todo Implement real INI file system.
bool_t main_file_exist(const char *p_name)
{
  FIL file;
  if (f_open(&file,p_name,FA_OPEN_EXISTING|FA_READ)!=FR_OK) return false;
  unsigned int bytes_read = 0;
  f_read(&file,bt_name,16,&bytes_read);
  bt_name[MIN(16,bytes_read)] = 0;
  f_close(&file);
  return true;
}


uint8_t Buffer[1024] __attribute__ ((aligned (4))) ;   /* Working buffer */
void put_dump(FIL *p_file, const uint8_t *p_data, uint32_t ofs, int cnt);
void main_sys_info(char *p_name);
void main_disk_info(char *p_name);

void put_dump(FIL *p_file, const uint8_t *p_data, uint32_t ofs, int cnt)
{
  int i;

  f_printf(p_file,"%08lX ",ofs);
  for (i=0; i<cnt; i++)
  {
    f_printf(p_file," %02X",p_data[i]);
  }
  f_putc(' ',p_file);
  for (i=0; i<cnt; i++)
  {
    if (p_data[i]<0x20 || p_data[i]>=0x7f) f_putc('.',p_file);
    else f_putc(p_data[i],p_file);
  }
  f_putc('\n',p_file);
}


void main_sys_info(char *p_name)
{
  FIL file;
  if (main_file_open(&file,p_name,FA_OPEN_ALWAYS|FA_WRITE)==NULL) return;

  uint8_t serial[8];
  ds18b20_read_serial(serial);
  int i;
  f_printf(&file,"1-wire serial number\t");
  for (i=0; i<8; i++)
  {
    f_printf(&file,"%02X",serial[i]);
  }
  f_printf(&file,"\n");

  int vpbdiv = VPBDIV&0x00000003;
  int pllcfg = PLLCFG&0x0000007f;
  f_printf(&file,"VPBDIV\t%02X\n",vpbdiv);
  f_printf(&file,"PLLCFG\t%02X\n",pllcfg);
  int m = (pllcfg&0x1f) + 1;
  int p = (pllcfg>>5)&0x03;
  f_printf(&file,"Fcco\t%ld Hz (m=%d, p=%d)\n",Fosc*m*2*p,m,p);

  long int fcclk = Fosc*m;
  f_printf(&file,"Fcclk\t%ld Hz\n",fcclk);

  if (vpbdiv==0) f_printf(&file,"Fpclk\t%ld Hz\n",fcclk/4);
  else if (vpbdiv==1) f_printf(&file,"Fpclk\t%ld Hz\n",fcclk);
  else if (vpbdiv==2) f_printf(&file,"Fpclk\t%ld Hz\n",fcclk/2);
  else f_printf(&file,"Illegal VPBDIV value\n");

  f_printf(&file,"MAMTIM\t%08X\n",MAMTIM);
  f_printf(&file,"MEMMAP\t%08X\n",MEMMAP);
  f_printf(&file,"CLOCKS_PER_SEC\t%d\n",CLOCKS_PER_SEC);

  f_close(&file);
  printf("Wrote file \"%s\"\n",p_name);
}


void main_disk_info(char *p_name)
{
  FIL file;
  if (main_file_open(&file,p_name,FA_OPEN_ALWAYS|FA_WRITE)==NULL) return;

  long l;
  uint8_t b;
  uint16_t w;

  if (disk_ioctl(0,GET_SECTOR_COUNT,&l)==RES_OK)
  {
    f_printf(&file,"Drive size: %lu sectors\n",l);
  }
  if (disk_ioctl(0,GET_SECTOR_SIZE,&w)==RES_OK)
  {
    f_printf(&file,"Sector size: %u\n",w);
  }
  if (disk_ioctl(0,GET_BLOCK_SIZE,&l)==RES_OK)
  {
    f_printf(&file,"Erase block size: %lu sectors\n",l);
  }
  if (disk_ioctl(0,MMC_GET_TYPE,&b)==RES_OK)
  {
    f_printf(&file,"MMC/SDC type: %u\n",b);
  }
  if (disk_ioctl(0,MMC_GET_CSD,Buffer)==RES_OK)
  {
    f_puts("CSD:\n",&file);
    put_dump(&file,Buffer, 0, 16);
  }
  if (disk_ioctl(0,MMC_GET_CID,Buffer)==RES_OK)
  {
    f_puts("CID:\n",&file);
    put_dump(&file,Buffer,0,16);
  }
  if (disk_ioctl(0,MMC_GET_OCR,Buffer)==RES_OK)
  {
    f_puts("OCR:\n",&file);
    put_dump(&file,Buffer,0,4);
  }
  if (disk_ioctl(0,MMC_GET_SDSTAT,Buffer)==RES_OK)
  {
    f_puts("SD Status:\n",&file);
    for (l=0; l<64; l+=16) put_dump(&file,Buffer+l,l,16);
  }

  f_close(&file);
  printf("Wrote file \"%s\"\n",p_name);
}


void main_log_data(const char *p_name_1, const char *p_name_2)
{
  FIL file1;
  FIL file2;
  FIL *p_file1 = main_file_open(&file1,p_name_1,FA_OPEN_ALWAYS|FA_WRITE);
  FIL *p_file2 = main_file_open(&file2,p_name_2,FA_OPEN_ALWAYS|FA_WRITE);

  if (p_file1==NULL && p_file2==NULL)
  {
    printf("Could not open log files. Logging canceled.\n");
    return;
  }

  int delay = 10;
  int samples = 0;
  uint32_t last_count = 0;

  printf("Logging");

  while (samples<300)
  {
    if (timer_get_10ms_counter()>=last_count+delay)
    {
      last_count = timer_get_10ms_counter();

      if (samples%60==0) printf("\n");
      samples += 1;
      uart0_putc('.');

      rtc_get_time(&rtc);
  
      float acc[3];
      accelerometer_read(acc);
      if (p_file1!=NULL)
      {
        f_printf(p_file1,"%02d\tX=%d\tY=%d\tZ=%d\n",rtc.seconds,(int)(1000.0*acc[0]),(int)(1000.0*acc[1]),(int)(1000.0*acc[2]));
      }

      float t = thermometer_read();
      if (p_file2!=NULL)
      {
        f_printf(p_file2,"%02d\tT=%d.%03dC\n",rtc.seconds,(int)t,(int)((t-(int)t)*1000.0));
      }
    }
  }

  f_close(&file1);
  f_close(&file2);

  printf("\ndone.\n");
}


int main(void)
{
  // System initialization.
  target_init();
  irq_vic_init();

  // Install a system timer running at 1 kHz.
  timer_init(SYSTEM_TIMER,Fpclk/1000 - 1);
  timer_enable(SYSTEM_TIMER);

  // Start the real time clock (at 1/1/2010).
  rtc.seconds = 0;
  rtc.minutes = 0;
  rtc.hours = 0;
  rtc.day_of_month = 1;
  rtc.day_of_week = 4;
  rtc.day_of_year = 1;
  rtc.month = 1;
  rtc.year = 2010;
  rtc_init();
  rtc_set_time(&rtc);
  rtc_start();

  // Start blinking an LED.
  led_init();
  timer_set_10ms_hook(main_10ms_timer);

  // Setup UART0 for 115200n81.
  uart0_init(115200,8,'n',1);

  // Initialize device driver system used for printf, the thermometer & accelerometer.
  /// \todo Move device_init to device drivers with check for multiple initializations.
  device_init();

  main_send_string(p_hello); // Send welcome message.
  main_send_string(p_version); // Show software version.
  timer_wait_ms(1500);  // Wait a bit.

  // Print instructions.
  printf("You can now spin, flip & rotate the Sceptre.\n");
  printf("During 30 seconds data will be logged to the SD card.\n");
  printf("Accelerometer data will be logged to \"%s\".\n",p_name1);
  printf("Thermometer data will be logged to \"%s\".\n",p_name2);
  printf("After logging a USB mass storage device will be started.\n");
  printf("Provide a file named \"%s\" to activate Bluetooth module.\n",p_bluetooth_test);
  printf("\n");
  timer_wait_ms(3500);  // Give the user some time to read the text.

  // Start thermometer device driver.
  if (thermometer_init()==false)
  {
    printf("thermometer_init failed.\n");
  }

  // Start accelerometer device driver.
  if (accelerometer_init()==false)
  {
    printf("accelerometer_init failed.\n");
  }

  // Try to install the file system.
  FATFS fs;
  bool_t sd_wp = false;
  if (sdcard_detect()==true)
  {
    printf("SD card detected.\n");
    sd_wp = sdcard_write_protected();
    if (sd_wp==true)
    {
      printf("SD card is write protected.\n");
      printf("Logging will be skipped.\n");
    }
    if (disk_initialize(0)==0)
    {
      // Mount file system.
      f_mount(0,&fs);
    }
    else
    {
      printf("Could not initialize disk.\n");
    }
  }

  if (sd_wp==false)
  {
    main_sys_info("sys_info.txt");
    main_disk_info("sd_info.txt");
  }

  bool_t disconnect = true;
  bool_t has_bluetooth = main_file_exist(p_bluetooth_test);
  if (has_bluetooth==true)
  {
    printf("Initializing Bluetooth module...\n");
    bluetooth_init(19200);
    bluetooth_power(BLUETOOTH_ON);
    timer_wait_ms(5000);  // Wait 5 seconds for module to stabilize.
    printf("   done\n");
    printf("Bluetooth waiting for connection (pin=1234).\n");
    printf("Friendly name is \"%s\"\n",bt_name);
    bluetooth_set_friendly_name(bt_name);
    //bluetooth_get_module_info();
    bluetooth_connect(true,"1234",0,100,true);
    printf("\n");
  }
  else
  {
    printf("Skipping Bluetooth test.\n");
  }

  if (sd_wp==false) main_log_data(p_name1,p_name2);  // Log some data.

  printf("Starting USB mass storage device.\n");
  usb_mass_storage_init();

  uint32_t last_count = 0;

  while (1)
  {
    usb_mass_storage_tick();

    if (timer_get_10ms_counter()>=last_count+100)
    {
      last_count = timer_get_10ms_counter();
      rtc_get_time(&rtc);

      float acc[3];
      accelerometer_read(acc);
      float t = thermometer_read();
      printf("%02d:%02d:%02d\tT=%0.4fC\tX=%f\tY=%f\tZ=%f\n",
             (int)rtc.hours,(int)rtc.minutes,(int)rtc.seconds,t,acc[0],acc[1],acc[2]);

      if (has_bluetooth==true)
      {
        if (bluetooth_is_connected()==true)
        {
          uint32_t ch;
          ch = 0;
          // Get data received on UART1 (if any).
          while (uart1_getc(&ch,0)==true)
          {
            uart0_putc(ch); // Echo on UART0 (Terminal).
          }
          if (ch!=0) printf("\n"); // New line for better visibility in the terminal.
          // Get data received on UART0 (if any).
          while (uart0_getc(&ch,0)==true)
          {
            uart1_putc(ch,true); // Echo on UART1 (Bluetooth).
          }
          disconnect = false;
        }
        else if (disconnect==false)
        {
          printf("Bluetooth disconnected\n");
          disconnect = true;
        }
      }
    }
  }

  return 0;
}
