/*****************************************************************************
*  accelero.c:  Accelerometer driver.
*
*  Author: Clemens Valens
*  License: GNU General Public License
*
*  History
*  2009.11.22  ver 1.00
* 
*  The accelerometer MMA7260Q is connected to AD0.1-3 (pins 13 to 15):
*  X = AD0.1 (P0.28, pin 13)
*  Y = AD0.2 (P0.29, pin 14)
*  Z = AD0.3 (P0.30, pin 15)
*  Sleep = P1.17 (active low)
*
******************************************************************************/

/**
	@file
	Low-level accelerometer driver implementation.
 */

#include "sceptre.h"
#include "accelero.h"


#define ADC_START_CONVERSION  AD0CR |= (1<<24)
#define ADC_STOP_CONVERSION  AD0CR &= 0xf8ffffff
#define ADC_CHANNEL_123  0x0000000e
#define ADC_GDR_DONE  (1<<31)


/**
* Initialize the accelerometer.
*
* @param adc_clock  Clock frequency of the A/D converter.
*/
void accelero_init(int adc_clock)
{
  // Set the related pins to ADC inputs, AD0.1-3
  PINSEL1 &= 0xc0ffffff;
  PINSEL1 |= 0x15000000;

  // CLKDIV = Fpclk/adc_clock - 1, PDN = 1 (normal operation).
  AD0CR = (((Fpclk/adc_clock-1)&0xff)<<8) | (1<<21);

  IODIR1 |= (1<<17);  // Make P1.17 an output.
  accelero_sleep(ACCELEROMETER_ON);
}


/**
* Control the accelerometer's sleep mode.
*
* @param sleep  ACCELEROMETER_OFF or ACCELEROMETER_ON
*/
void accelero_sleep(int sleep)
{
  if (sleep==ACCELEROMETER_OFF) IOCLR1 = (1<<17);  // Make P1.17 low.
  else IOSET1 = (1<<17);  // Make P1.17 high.
}


/**
* Read one channel of the accelerometer.
*
* @param channel  Accelerometer channel to read.
* @return The value of the accelerometer channel.
*/
float accelero_read_one(int channel)
{
  AD0CR &= 0xffffff00;
  AD0CR |= (1<<channel); // Select channel.

  ADC_START_CONVERSION;  // Start ADC.
  // Wait for conversion to finish.
  while ((AD0GDR&ADC_GDR_DONE)==0);
  ADC_STOP_CONVERSION;  // Stop ADC.

  return ((AD0GDR>>6)&0x3ff)/1024.0;
}


/**
* Read all channels of the accelerometer.
*
* @param p_dst  Address to write the accelerometer data to.
* @return The number of bytes written.
*/
int accelero_read(float *p_dst)
{
  for (int i=0; i<3; i++)
  {
    p_dst[i] = accelero_read_one(i+1);
  }
  return 3*sizeof(float);
}

