/****************************************************************************
*  accelero_device.c: accelerometer device driver.
*
*  This is a simple example for a device driver that allows the use of
*  standard C file functions like fopen, fclose & fread to control it.
*
*  Author: Clemens Valens
*  License: GNU General Public License
*
*  History
*  2009.11.22  ver 1.00
*
*****************************************************************************/

/**
	@file
	Accelerometer device driver implementation.
 */

#include "sceptre.h"
#include "device.h"
#include "accelero.h"
#include <stdlib.h>  // For atoi.


#ifdef ACCELERO_DEBUG
#include <stdio.h>
#endif


int accelero_dev_open(struct _reent *p_reent, const char *p_name, int flags, int mode);
_ssize_t accelero_dev_read(struct _reent *p_reent, int file, void *p_dst, size_t dst_size);
_ssize_t accelero_dev_close(struct _reent *p_reent, int file);


/**
* A pointer to this structure has to be registered in the device_table_list.
*/
device_table_entry_t accelero_device =
{
  "accelero",             // Device "file" name (without forward slashes).
  SCEPTRE_ACCELEROMETER,  // Device ID.
  accelero_dev_open,      // open
  accelero_dev_close,     // close
  accelero_dev_read,      // read
  NULL,                   // write
  NULL,                   // init
  NULL                    // ioctl
};


/**
* Open function for the accelerometer device driver.
*
* This function gets called when you do something like this:
*
*   FILE *p_accelerometer = fopen("/accelero/","rw");
*
* Note the forward slashes surrounding the "file" name. Everything that
* follows the string "/accelero/" is passed to this function in p_name, i.e.
*
*   FILE *p_accelerometer = fopen("/accelero/1000000","rw");
*
* will pass "1000000" (adc clock frequency) in p_name to this function.
*
* @param p_reent  Pointer to a structure that will receive the return value.
* @param p_name   Pointer to the name of the file to open.
* @param flags    Flags.
* @param mode     Mode.
* @return Valid handle (never fails).
*/
int accelero_dev_open(struct _reent *p_reent, const char *p_name, int flags, int mode)
{
#ifdef ACCELERO_DEBUG
  printf("accelero_dev_open(%s,%d,%d)\n",p_name,flags,mode);
#endif
  // Open the device.
  accelero_init(atoi(p_name));
  return DEVICE(SCEPTRE_ACCELEROMETER);
}


/**
* Read function for the accelerometer device driver.
*
* This function gets called when you do something like this after
* obtaining a valid FILE handle with fopen():
*
*   float data[3];
*   fread(data,sizeof(float),3,p_accelerometer);
*
* @param p_reent   Pointer to a structure that will receive the return value.
* @param file      Handle of the file to read from.
* @param p_dst     Address to write the data to.
* @param dst_size  Number of bytes to read.
* @return The number of bytes read on success, -1 otherwise.
*/
_ssize_t accelero_dev_read(struct _reent *p_reent, int file, void *p_dst, size_t dst_size)
{
#ifdef ACCELERO_DEBUG
  printf("accelero_dev_read(%08x,%d,%08x,%d)\n",p_reent,file,p_dst,dst_size);
#endif

  int size = accelero_read(p_dst);

#ifdef ACCELERO_DEBUG
  for (int j=0; j<size/sizeof(float); j++) printf("%f ",((float*)p_dst)[j]);
#endif

  if (size<0) p_reent->_errno = ENODEV; // Set reason of failure.
  return size;
}


/**
* Close function for the accelerometer device driver.
*
* This function gets called when you do something like this after
* obtaining a valid FILE handle with fopen():
*
*   fclose(p_accelerometer);
*
* @return Always 0.
*/
_ssize_t accelero_dev_close(struct _reent *p_reent, int file)
{
  accelero_sleep(ACCELEROMETER_OFF);
  return 0;
}
