/*****************************************************************************
 *   bluetooth.c:  Bluetooth API for Sceptre
 *
 *   Author: Clemens Valens
 *   License: GNU General Public License 
 *
 *   History
 *   2009.11.05  ver 1.00    Preliminary version, first release
 *
******************************************************************************/

/**
	@file
	High-level Bluetooth driver implementation.
 */

#include "sceptre.h"
#include "bluetooth.h"
#include "btm222.h"  // The driver for the Bluetooth module.

/// Bluetooth module power control is on pin P1.18
#define BLUETOOTH_PIN  18
#define BLUETOOTH_POWER_OFF  IOSET1 = (1<<BLUETOOTH_PIN)
#define BLUETOOTH_POWER_ON  IOCLR1 = (1<<BLUETOOTH_PIN)


/**
* Switch module power on or off.
*
* @param power    BLUETOOTH_OFF or BLUETOOTH_ON. 
*/
void bluetooth_power(int power)
{
  if (power==BLUETOOTH_OFF) BLUETOOTH_POWER_OFF;
  else BLUETOOTH_POWER_ON;
}


/**
* Initialize the Bluetooth driver.
*
* @param baudrate    Speed of the serial port that talks to the BT module. 
* @return true if the initialization went well, false otherwise.
*/
bool_t bluetooth_init(int baudrate)
{
  IODIR1 |= (1<<BLUETOOTH_PIN);  // Make pin an output.
  bluetooth_power(BLUETOOTH_OFF);
  return uart1_init(baudrate,8,'n',1);
}


/**
* Set the human readable name of the Bluetooth device.
*
* @param p_name   Friendly name as ASCII string (16 characters max).
* @return true if name was set, false otherwise.
*/
bool_t bluetooth_set_friendly_name(char *p_name)
{
  return btm222_command("N=",p_name,BTM222_MAX_PEER_ID);
}


/**
* Set the pin (ID) code of the Bluetooth device.
*
* @param p_pin   Pin (ID) as ASCII string (4 characters max).
* @return true if name was set, false otherwise.
*/
bool_t bluetooth_set_pin(char *p_pin)
{
  return btm222_command("P=",p_pin,4);
}


/**
* Send data over the Bluetooth connection.
*
* @param p_data     Pointer to the data to send. 
* @param data_size  Number of bytes to send.
* @return true if a connection was established, false otherwise.
*
* @todo  Implement. While doing that, implement also a receive function.
*/
int bluetooth_send(char *p_data, int data_size)
{
  return 0;
}


/**
* Do a search for other Bluetooth devices.
*
* @return the number of devices found.
*
* @todo  Make more useful.
*/
uint8_t bluetooth_scan(void)
{
  return btm222_seek_devices();
}


/**
* Try to connect to another Bluetooth device.
*
* @param as_slave    true for connecting as slave (passive), false for master. 
* @param p_pin       ID as ASCII string.
* @param device      Number of the device (1..8) to connect to (master mode only).
* @param timeout_ms  Time (ms) to wait (block) for a connection.
* @param reconnect   Keep a background connection "thread" running (slave mode only).
* @return true if a connection was established, false otherwise.
*
* @todo  Implement master mode.
*/
bool_t bluetooth_connect(bool_t as_slave, char *p_pin, uint8_t device, uint32_t timeout_ms, bool_t reconnect)
{
  if (as_slave==true)
  {
    btm222_write_command((uint8_t*)BTM_CMD_ROLE_SLAVE);
    btm222_write_command((uint8_t*)BTM_CMD_CONNECT_AUTO);
  }
  else
  {
    btm222_write_command((uint8_t*)BTM_CMD_ROLE_MASTER);
    btm222_write_command((uint8_t*)BTM_CMD_CONNECT_NOAUTO);
  }
  bluetooth_set_pin(p_pin);
  return btm222_wait_for_connect(timeout_ms,reconnect);
}


/**
* Check if this device is connected.
*
* @return true if a connection exists, false if not.
*/
bool_t bluetooth_is_connected(void)
{
  return btm222_is_connected();
}


/**
* Get some information concerning the Bluetooth module.
*
* @return true if a connection exists, false if not.
*
* @todo  Format the module info in some structure.
*/
void bluetooth_get_module_info(void)
{
  btm222_get_all_parameters();
}
