/**
	@file
	RDCF2 wrapper.
 */

#include <stdio.h>
#include <string.h>
#include "type.h"
#include "rdcf2.h"
#include "ssp.h"
#include "sdcard_driver.h"
#include "drive.h"


#define SIZEOF_DIR_ENTRY  32

extern struct rdcf  fcbs [MaxFileBuffers];

/*******************************************************************
 * structure defs
*******************************************************************/

struct PARTITION_ENTRY
{
  // single partition entry for MMC.
  uint8 BootActive;
  uint FirstPartitionSector:24;
  uint8 FileSystemDescriptor;
  uint LastPartitionSector:24;
  uint32 FirstSectorPosition;
  uint32 NumberSectorsInPartition;
} __attribute__ ((packed));


struct PARTITION_TABLE
{
  uint8 BootCode[446];
  struct PARTITION_ENTRY  PartitionEntry0;
  struct PARTITION_ENTRY  PartitionEntry1;
  struct PARTITION_ENTRY  PartitionEntry2;
  struct PARTITION_ENTRY  PartitionEntry3;
  uint16  Signature;
} __attribute__ ((packed));


struct BOOTSECTOR_ENTRY
{
  // this description is for an MMC boot block not for MSDOS per se.
  uint JumpCommand:24;
  int8 OEM_NAME[8];
  uint16 BytesPerSector;
  uint8 SectorsPerCluster;
  uint16 ReservedSectors;
  uint8 NumberOfFATs;
  uint16 NumberRootDirEntries;
  uint16 NumberOfSectorsOnMedia;
  uint8 MediaDescriptor;
  uint16 SectorsPerFAT;
  uint16 SectorsPerTrack;
  uint16 NumberOfHeads;
  uint32 NumberOfHiddenSectors;
  uint32 NumberOfTotalSectors;
  uint8 DriveNumber;
  uint8 __RESERVED__;
  uint8 ExtendedBootSignature;
  uint32 VolumeID;
  int8 VolumeLabel[11];
  int8 FileSystemType[8];
  uint8 LoadProgramCode[448];
  uint16 Signature;
} __attribute__ ((packed));

struct DATABUFFER
{
  uint8 data [RDCF_SECTOR_SIZE];
};

  // re-use buffer area, don't waste RAM.
union MMC_IO_BUFFER
{
  struct BOOTSECTOR_ENTRY   BootBlock;
  struct PARTITION_TABLE    PartitionTable;
};

/*******************************************************************
 * structure vars
*******************************************************************/
union MMC_IO_BUFFER IoBuffer;   // scratch area for transitory file i/o.
DRIVE_DESCRIPTION DriveDesc;


/*******************************************************************
 * routines specific to rdcf2 operation.
*******************************************************************/
static void drive_init_rdcf2_struct(void);
static void drive_collect_data(void);

static void drive_init_rdcf2_struct(void)
{
  int i;
  // Remember to do the reserved fcb as well.
  for (i=0; i<MaxFileBuffers+1; i++)
  {
    fcbs[i].ReadSector = sdcard_read_block;
    fcbs[i].WriteSector = sdcard_write_block;
    fcbs[i].BufferInUse = FALSE;
  }
}


static void drive_collect_data(void)
{
  // How large are the FAT tables?
  DriveDesc.SectorsPerFAT = IoBuffer.BootBlock.SectorsPerFAT;
  // Important info to decode FAT entries into sectors.
  DriveDesc.SectorsPerCluster = IoBuffer.BootBlock.SectorsPerCluster;
  // First File Allocation Table.
  DriveDesc.FirstFatSector = DriveDesc.SectorZero + IoBuffer.BootBlock.ReservedSectors;
  // "backup" copy of the First FAT. usually to undelete files.
  if (IoBuffer.BootBlock.NumberOfFATs>1)
  {
    DriveDesc.SecondFatSector = DriveDesc.FirstFatSector + DriveDesc.SectorsPerFAT;
  }
  else
  {
    // nope, only one FAT...
    DriveDesc.SecondFatSector = -1;
  }
  // Where does the actual drive data area start?
  if (DriveDesc.SecondFatSector == -1)
  {
    // only one FAT, so data follows first FAT.
    DriveDesc.RootDirSector = DriveDesc.FirstFatSector + IoBuffer.BootBlock.SectorsPerFAT;
  }
  else
  {
    // data follows both FAT tables.
    DriveDesc.RootDirSector = DriveDesc.FirstFatSector + (2 * IoBuffer.BootBlock.SectorsPerFAT);
  }
  // How many entries can be in the root directory?
  DriveDesc.NumberRootDirEntries = IoBuffer.BootBlock.NumberRootDirEntries;
  // where does cluster 2 begin?
  DriveDesc.DataStartSector = DriveDesc.RootDirSector +
    (DriveDesc.NumberRootDirEntries * SIZEOF_DIR_ENTRY) / RDCF_SECTOR_SIZE;
  // Where does the partition end?
  DriveDesc.MaxDataSector = DriveDesc.SectorZero +
    ((IoBuffer.BootBlock.NumberOfSectorsOnMedia) ?
    IoBuffer.BootBlock.NumberOfSectorsOnMedia :
    IoBuffer.BootBlock.NumberOfTotalSectors);
}

/********************************************************************
 * drive_init
 * return TRUE if error, FALSE if everything went well.
********************************************************************/

BOOL drive_init(void)
{
  // Configure IO pins.
  sdcard_hardware_init();
  // Power on SD card.
  sdcard_power(SDCARD_POWER_ON);

  // Initialize the SSP interface.
  ssp_init();

  // Access drive and collect structure info.
  // See if we have a card inserted.
  if (sdcard_detect()==FALSE)
  {
    // drive disappeared!
    DriveDesc.IsValid = FALSE;
    return TRUE;
  }
  if (DriveDesc.IsValid==TRUE)
  {
    // we already know about this drive.
    return FALSE;
  }
  if (sdcard_initialise()==FALSE) return TRUE; // return TRUE on error, remember?
  // init fcbs.
  drive_init_rdcf2_struct();
  // get the partition table to find the boot block.
  if (sdcard_read_block(0,(uint8*)&IoBuffer)) return TRUE;
  // validate.
  if (IoBuffer.PartitionTable.Signature != 0xaa55) return TRUE;
  // get the boot block now.
  DriveDesc.SectorZero = IoBuffer.PartitionTable.PartitionEntry0.FirstSectorPosition;
  if (sdcard_read_block(DriveDesc.SectorZero,(uint8*)&IoBuffer)) return TRUE;
  // validate.
  if (IoBuffer.BootBlock.Signature != 0xaa55) return TRUE;
  // looks good, make a note of where stuff starts at.
  drive_collect_data();
  // pass all tests before validating drive.
  DriveDesc.IsValid = TRUE;
  return FALSE;
}

