/****************************************************************************
*  sdcard_device.c: SD card device driver.
*
*  This is a simple example for a device driver that allows the use of
*  standard C file functions like fopen, fread & fwrite to control it.
*
*  Author: Tom Walsh tom@openhardware.net
*  License: GNU General Public License
*
*  History
*  2009.12.10  ver 1.00
*
*****************************************************************************/

/**
	@file
	SD card device driver for use with FatFs.

  \todo Get it finished & working.
 */

#include "sceptre.h"
#include <stdio.h>
#include "ff.h" // This is the actual file system.
#include "device.h"
#include "sdcard_device_ff.h"
#include "drive.h"


static uint8_t sdcard_allocate_fcb(void);
static int sdcard_open(struct _reent *p_reent, const char *p_name, int flags, int always666);
static int sdcard_close(struct _reent *p_reent, int file);
static _ssize_t sdcard_read(struct _reent *p_reent, int file, void *p_dst, size_t dst_size);
static _ssize_t sdcard_write(struct _reent *p_reent, int file, const void *p_src, size_t src_size);
static int ioctl_dos_seek(struct _reent *p_reent, int file, _off_t pos, int whence);
static int ioctl_dos_unlink(struct _reent *p_reent, char *p_name);
static int ioctl_dos_rename(struct _reent *p_reent, const char *p_old, const char *p_new);
static int ioctl_dos_flush_dir(struct _reent *p_reent, int file);
static int sdcard_ioctl(struct _reent *p_reent, int file, int cmd, void *ptr);
static int sdcard_init(void);


device_table_entry_t sdcard_device =
{
  "sdcard",        // Device "file" name (without forward slashes).
  SCEPTRE_SDCARD,  // Device ID.
  sdcard_open,     // open
  sdcard_close,    // close
  sdcard_read,     // read
  sdcard_write,    // write
  sdcard_init,     // init
  sdcard_ioctl     // ioctl
};


extern struct DRIVE_DESCRIPTION Drive;

// Note, last file buffer is reserved
// for special ops such as delete, rename
// dir listings etc.
FIL fcbs[MaxFileBuffers+1];


static uint8_t sdcard_allocate_fcb(void)
{
  // Find free fcb and return it or 0 if none available.
  int i;
  for (i=0; i<MaxFileBuffers; i++)
  {
    if (fcbs[i].BufferInUse!=0) continue;
    return i;
  }
  return -1;
}


static int sdcard_open(struct _reent *p_reent, const char *p_name, int flags, int always666)
{
  int result;
  unsigned handle;

  // Is a drive still there?
  if (DriveDesc.IsValid==0)
  {
    p_reent->_errno = ENODEV;
    return -1;
  }

  // Find a buffer to use.
  if ((handle=sdcard_allocate_fcb())==-1)
  {
    p_reent->_errno = ENOBUFS;
    return -1;
  }

  result = f_open(&fcbs[handle],p_name,flags);
  if (result!=0)
  {
    p_reent->_errno = ~result;
    return -1;
  }

  return DEVICE(SCEPTRE_SDCARD) | handle;
}


static int sdcard_close(struct _reent *p_reent, int file)
{
  int result;

  // Is a drive still there?
  if (DriveDesc.IsValid==0)
  {
    p_reent->_errno = ENODEV;
    return -1;
  }

  result = f_close(&fcbs[file&0xff]);
  if (result!=0)
  {
    p_reent->_errno = ~result;
    return -1;
  }

  return 0;
}


static _ssize_t sdcard_read(struct _reent *p_reent, int file, void *p_dst, size_t dst_size)
{
  int result;

  // Is a drive still there?
  if (DriveDesc.IsValid==0)
  {
    p_reent->_errno = ENODEV;
    return -1;
  }

  result = f_read(&fcbs[file&0xff],p_dst,dst_size);
  if (result<0)
  {
    p_reent->_errno = ~result;
    return -1;
  }

  return result;
}


static _ssize_t sdcard_write(struct _reent *p_reent, int file, const void *p_src, size_t src_size)
{
  int result;

  // Is a drive still there?
  if (DriveDesc.IsValid==0)
  {
    p_reent->_errno = ENODEV;
    return -1;
  }

  result = f_write(&fcbs[file&0xff],p_src,src_size);
  if (result<0)
  {
    p_reent->_errno = ~result;
    return -1;
  }

  return result;
}


static int ioctl_dos_seek(struct _reent *p_reent, int file, _off_t pos, int whence)
{
  int result;

  switch (whence)
  {
    case SEEK_SET:
      result = f_lseek(&fcbs[file&0xff],pos);
      if (result<0)
      {
        p_reent->_errno = ~result;
        return -1;
      }
      return fcbs[file & 0xff].position;

    case SEEK_CUR:  // not implemented.
    case SEEK_END:  // not implemented.
      break;
  }
  p_reent->_errno = EINVAL;
  return -1;
}


static int ioctl_dos_unlink(struct _reent *p_reent, char *p_name)
{
  int result = f_unlink(p_name);
  if (result<0)
  {
    p_reent->_errno = ~result;
    return -1;
  }
  return 0;
}


static int ioctl_dos_rename(struct _reent *p_reent, const char *p_old, const char *p_new)
{
  int result = f_rename(p_old,p_new);
  if (result<0)
  {
    p_reent->_errno = ~result;
    return -1;
  }
  return 0;
}


/*static int ioctl_dos_flush_dir(struct _reent *p_reent, int file)
{
  // Flush dir entry associated with file.
  int result;
  if ((file&0xff)<0 || (file&0xff)>=MaxFileBuffers)
  {
    return -1;
  }

  result = rdcf_flush_directory(&fcbs[file&0xff]);
  if (result<0)
  {
    p_reent->_errno = ~result;
    return -1;
  }
  return 0;
}*/


static int sdcard_ioctl(struct _reent *p_reent, int file, int cmd, void *ptr)
{
  // Is a drive still there?
  if (DriveDesc.IsValid==0)
  {
    p_reent->_errno = ENODEV;
    return -1;
  }

  switch (cmd)
  {
    case SDCARD_IOCTL_SEEK:
      // ptr is to a structure containing two pointers to pointers.
      return ioctl_dos_seek(p_reent,file,*(_off_t*)((long*)ptr)[0],*(int*)((long*)ptr)[1]);

    case SDCARD_IOCTL_UNLINK:
      // ptr is the address of a pointer to filename string.
      return ioctl_dos_unlink(p_reent,(char*)((long*)ptr)[0]);

    case SDCARD_IOCTL_RENAME:
      return ioctl_dos_rename(p_reent,(const char*)((long*)ptr)[0],(const char*)((long*)ptr)[1]);

    //case SDCARD_IOCTL_FLUSH_DIR:
    //  return ioctl_dos_flush_dir(p_reent,file);
  }

  // Anything not implemented is "INVALID".
  p_reent->_errno = EINVAL;
  return -1;
}


static int sdcard_init(void)
{
  // Init the drive system & software.
  return drive_init();
}


