/******************************************************
 *
 * MMC interface routines.
 *
 * (C) 2005 - Tom Walsh tom@openhardware.net
 *
 *
******************************************************/

/**
	@file
	MMC interface routines.
 */

//#include <time.h>
#include "sceptre.h"
#include "sdcard_driver.h"
#include "ssp.h"
#include "rdcf2.h"


#define SDCARD_POWER_BIT  (16)  /* P1.16 output - SD card powered if low. */
#define SDCARD_WRITE_PROTECT_BIT  (24)  /* P1.24 input - SD card write protected if low. */
#define SDCARD_DETECT_BIT  (25)  /* P1.25 input - SD card detected if low. */
#define SDCARD_CS_BIT  (20)  /* P0.20 output - SD card chip select. */
// #define SDCARD_SCK_BIT  (17)     // P0.17 SSP to SD card.
// #define SDCARD_MISO_BIT  (18)     // P0.18 MISO from SD card.
// #define SDCARD_MOSI_BIT  (19)     // P0.19 MOSI to SD card.


#define sdcard_status_idle				  (1<<0)	// in IDLE state and initializing process.
#define sdcard_status_erase_reset		(1<<1)	// erase sequence aborted.
#define sdcard_status_illegal_cmd		(1<<2)	// illegal command was sent.
#define sdcard_status_error_crc			(1<<3)	// bad CRC in last command.
#define sdcard_status_erase_seq_err	(1<<4)	// error in sequence of erase commands.
#define sdcard_status_addr_error		(1<<5)	// address did not match block length.
#define sdcard_status_param_error		(1<<6)	// illegal argument used in command.
#define sdcard_status_data_accepted		0x5		// write command is ok to go.
#define sdcard_status_data_crc_error		0xb		// crc error on data block.
#define sdcard_status_data_write_error	0xd		// write error on media (?).


#define SDCARD_CMD0  0x40	// software reset
#define SDCARD_CMD1  0x41	// brings card out of idle state
#define SDCARD_CMD13  0x4d	// read status word of card.
#define SDCARD_CMD16  0x50	// sets the block length used by the memory card
#define SDCARD_CMD17  0x51	// read single block
#define SDCARD_CMD24  0x58	// writes a single block

#define SDCARD_MAX_CMD  8
uint8_t sdcard_cmd[SDCARD_MAX_CMD];

uint8_t sdcard_cmd_reset[] = { 0x40, 0x00, 0x00, 0x00, 0x00, 0x95 };
uint8_t sdcard_cmd_init[] = { 0x41, 0x00, 0x00, 0x00, 0x00, 0xff };
uint8_t sdcard_cmd_set_block[] = { SDCARD_CMD16, 0, 0, 2, 0, 0xff };



/******************************************************
 *
 * sdcard_select and sdcard_unselect
 *
 * board specific routines to communicate with SD card.
 *
******************************************************/


void sdcard_hardware_init(void)
{
  IODIR1 |= (1<<SDCARD_POWER_BIT); // Output.
  IODIR1 &= ~(1<<SDCARD_WRITE_PROTECT_BIT); // Input.
  IODIR1 &= ~(1<<SDCARD_DETECT_BIT); // Input.
}


void sdcard_power(uint8_t power)
{
  if (power==SDCARD_POWER_ON) IOCLR1 = (1<<SDCARD_POWER_BIT);
  else IOSET1 = (1<<SDCARD_POWER_BIT);
}


bool_t sdcard_detect(void)
{
  return (IOPIN1&(1<<SDCARD_DETECT_BIT))==0? true : false;
}


bool_t sdcard_write_protected(void)
{
  return (IOPIN1&(1<<SDCARD_WRITE_PROTECT_BIT))==0? false : true;
}


void sdcard_select(void)
{
  // Select SSP target.
  IOCLR0 = (1<<SDCARD_CS_BIT);  // set SS = 0 (on)
}


void sdcard_unselect(void)
{
  // Unselect SSP target.
  IOSET0 = (1<<SDCARD_CS_BIT);  // set SS = 1 (off)
}


/******************************************************
 *
 * sdcard_status
 *
 * Get response status byte and see if it matches
 * what we are looking for.
 * Return true if we DID get what we wanted.
 *
******************************************************/
bool_t sdcard_status(uint8_t response)
{
  int count = 4000;
  uint8_t result = ~response;
  while (result!=response && --count>0) result = ssp_get_byte();
  return (count!=0);  // Loop was exited before timeout?
}


/******************************************************
 *
 * sdcard_initialise
 *
 * condition SD card for operation
 *
 * Returns true if all went well or false if not good.
 *
******************************************************/
bool_t sdcard_initialise(void)
{
  unsigned int count;
  bool_t  result;

  sdcard_unselect();
  for (count=0; count<10; count++)
  {
    ssp_put_byte(0xFF);
  }

  sdcard_select();
  ssp_send(sdcard_cmd_reset,6);
  if (sdcard_status(sdcard_status_idle)==false)
  {
    sdcard_unselect();
    ssp_get_byte();
    return false;
  }

  count = 255;
  do
  {
    ssp_send(sdcard_cmd_init,6);
    count--;
  }
  while ((!sdcard_status(0x00)) && (count>0));
  sdcard_unselect();
  ssp_get_byte();
  if (count==0) return false;

  sdcard_select();
  ssp_send(sdcard_cmd_set_block,6);
  result = sdcard_status(0x00);
  sdcard_unselect();
  ssp_get_byte();

  return result;
}

/******************************************************
 *
 * sdcard_read_block
 *
 * called with:
 *  sector number to read, target buffer addr
 *  offset into sector and number of bytes read.
 *
 * return true on error, false if all went well.
 *
******************************************************/
bool_t sdcard_read_block(long sector, uint8_t *p_data)
{
  sector <<= 1;
  sdcard_select();
  sdcard_cmd[0] = SDCARD_CMD17;
  sdcard_cmd[1] = (sector>>16) & 0xff;
  sdcard_cmd[2] = (sector>>8) & 0xff;
  sdcard_cmd[3] = sector & 0xff;
  sdcard_cmd[4] = 0;
  sdcard_cmd[5] = 0xff;
  ssp_send(sdcard_cmd,6);
  if (sdcard_status(0x00))
  {
    // Get the data token for single block read.
    if (sdcard_status(0xfe))
    {
      ssp_read(p_data,RDCF_SECTOR_SIZE);
      // Read off, and discard CRC bytes.
      ssp_get_byte();
      ssp_get_byte();
    }
    else
    {
      sdcard_unselect();
      ssp_get_byte();
      return true;
    }
  }
  else
  {
    sdcard_unselect();
    ssp_get_byte();
    return true;
  }
  sdcard_unselect();
  ssp_get_byte();
  return false;
}

/******************************************************
 *
 * sdcard_get_write_result
 *
 * MMC will be busy (0xff), wait until status code
 * comes back and return that
 *
******************************************************/
uint8_t sdcard_get_write_result (void)
{
  int count = 60000l;
  uint8_t result = 0;
  while (result==0 && --count)
  {
    result = ssp_get_byte();
  }
  return result;
}


/******************************************************
 *
 * sdcard_write_block
 *
 * called with:
 *   sector to write and source buffer.
 * this always writes RDCF_SECTOR_SIZE chunks of one sector
 *
 * returns true on error, false if all went well.
 *
******************************************************/

bool_t sdcard_write_block(long sector, const uint8_t *p_data)
{
  int result = 0;
  sector <<= 1;
  sdcard_select();
  sdcard_cmd[0] = SDCARD_CMD24;
  sdcard_cmd[1] = (sector>>16) & 0xff;
  sdcard_cmd[2] = (sector>>8) & 0xff;
  sdcard_cmd[3] = sector & 0xff;
  sdcard_cmd[4] = 0;
  sdcard_cmd[5] = 0xff;
  ssp_send (sdcard_cmd, 6);
  
  if (sdcard_status(0x00))
  {
    // Send data token for single block write.
    ssp_put_byte(0xfe);
    ssp_send(p_data,RDCF_SECTOR_SIZE);
    // Dummy the CRC.
    ssp_put_byte(0xff);
    ssp_put_byte(0xff);
    // Next we see if all went well.
    result = ssp_get_byte();
    if ((result&0xf)!=sdcard_status_data_accepted)
    {
      // Something went wrong with block write.
      sdcard_unselect();
      ssp_get_byte();
      return true;
    }
  }
  else
  {
    // Failed to get "ok" for SDCARD_CMD24.
    sdcard_unselect();
    ssp_get_byte();
    return true;
  }
  
  // Wait for operation to complete itself.
  // We wait until the card is no longer busy.
  result = sdcard_get_write_result();
  
  // No longer busy, proceed.
  sdcard_unselect();
  ssp_get_byte();
  return false;
}

