/*-----------------------------------------------------------------------*/
/* MMC/SDSC/SDHC (in SPI mode) control module for LPC2xxx SSP            */
/* (C) Martin Thomas, 2009 - based on the AVR module (C)ChaN, 2007       */
/*-----------------------------------------------------------------------*/

/* Copyright (c) 2009, Martin Thomas, ChaN
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.
   * Neither the name of the copyright holders nor the names of
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. */

/*-----------------------------------------------------------------------*/
/*
 * Martin Thomas 5/2009
 * - inspired by the EFSL interface-driver by Mike Anton and me
 * - using ChaN's driver-skeleton from his AVR mmc.c
 * - This code is for the SSP-Interface ("SPI1") and fast-IO (FIO).
 *   It is not prepared for "old style" SPI or "legacy/slow" IO
 * - Olimex LPC-P2148: P0.17 SD-Card clock, P0.18 from SD-Card,
     P0.19 to SD-Card, P0.20 SD-Card select - NO ext. pull-up mounted
     P1.24 write protect switch (ext. pull-up, trace must be turned off),
     P1.25 card-detect switch (ext. pull-up, trace must be turned off)
*/
/*-----------------------------------------------------------------------*/

/**
	@file
	Mixed SD card/SPP driver.

  \todo Split in separate files (spp, mmc & sdcard).
 */

#define USE_FIFO  1

#include "sceptre.h"


/*-----------------------------------------------------------------------*/
/* SPI-Interface (Platform dependent, mthomas)                           */
/*-----------------------------------------------------------------------*/

#define SSPCR0_DSS      0
#define SSPCR0_CPOL     6
#define SSPCR0_CPHA     7
#define SSPCR0_SCR      8
#define SSPCR1_SSE      1
#define SSPSR_TNF       1
#define SSPSR_RNE       2
#define SSPSR_BSY       4

#define PCONP_PCSPI1    10

#define SD_CS_BIT       20
#define SD_CS_DIR       IODIR0 /*FIO0DIR*/
#define SD_CS_SET       IOSET0 /*FIO0SET*/
#define SD_CS_CLR       IOCLR0 /*FIO0CLR*/
#define SD_CS_SEL       PINSEL1
#define SD_CS_SEL_BIT   8

// bit-frequency = PCLK / (CPSDVSR * [SCR+1]), here SCR=0, PCLK=60MHz, must be even
#define SPI_SPEED_20MHz		4	/* => 15MHz */
#define SPI_SPEED_25MHz		4	/* => 15MHz */
#define SPI_SPEED_400kHz  150	/* => 400kHz */


// select card ( MMC CS = L )
static inline void SELECT(void)
{
	SD_CS_CLR = ( 1UL << SD_CS_BIT );
}

// unselect card ( MMC CS = H )
static inline void DESELECT( void )
{
	SD_CS_SET = ( 1UL << SD_CS_BIT );
}


static void spiSetSpeed(BYTE speed)
{
	speed &= 0xFE;
 	if (speed<2) speed = 2; 	
  SSPCPSR = speed;
}

/* general SPI send/receive */
static inline BYTE spi_write_read(BYTE outgoing)
{
	BYTE incoming;

	/// while( !( SSPSR & (1 << SSPSR_TNF ) ) ) { ; }
	SSPDR = outgoing;
	while ((SSPSR&(1<<SSPSR_RNE))==0) 
  { 
  }
	incoming = SSPDR;

	return incoming;
}


#if USE_FIFO
/* on LPC2k the FIFOs have 8 elements which each can hold up to 16bits */
#define FIFO_ELEM 8
/*
	BYTE *buff,  Data buffer to store received data
	UINT btr     Byte count (must be multiple of 4)
*/
static inline void spi_rcvr_block(BYTE *buff, UINT btr)
{
	UINT hwtr, startcnt, i, rec;

	hwtr = btr/2;
	if (btr<FIFO_ELEM) 
  {
		startcnt = hwtr;
	} 
  else 
  {
		startcnt = FIFO_ELEM;
	}

	SSPCR0 |= 0x0000000f;  // DSS to 16 bit

	for (i=startcnt; i; i--) 
  {
		SSPDR = 0xffff;  // fill TX FIFO
	}

	do 
  {
		while ((SSPSR&(1<<SSPSR_RNE))==0) 
    {
			// wait for data in RX FIFO (RNE set)
		}
		rec = SSPDR;
		if (i<(hwtr-startcnt)) 
    {
			SSPDR = 0xffff;
		}
		*buff++ = (BYTE)(rec>>8);
		*buff++ = (BYTE)(rec);
		i++;
	} 
  while (i<hwtr);

	SSPCR0 &= ~0x00000008;  // DSS to 8 bit
}

/*
	const BYTE *buff	 512 byte data block to be transmitted.
*/
static inline void spi_xmit_block(const BYTE *buff)
{
	UINT cnt;
	WORD data;

	SSPCR0 |= 0x0000000f;  // DSS to 16 bit

	for (cnt=0; cnt<(512/2); cnt++) 
  {
		while ((SSPSR&(1<<SSPSR_TNF))==0) 
    {
			// wait for TX FIFO not full (TNF)
		}
		data  = (*buff++) << 8;
		data |= *buff++;
		SSPDR = data;
	}

	while ((SSPSR&(1<<SSPSR_BSY))!=0)
  {
		// wait for BSY gone
	}
	while ((SSPSR&(1<<SSPSR_RNE))!=0)
  {
		data = SSPDR; // drain receive FIFO
	}

	SSPCR0 &= ~0x00000008;  // DSS to 8 bit
}
#endif /* USE_FIFO */


static void init_spi( void )
{
	UINT dummy;

	// configure CS as GPIO output and deselect card
	SD_CS_SEL = (SD_CS_SEL & ~(3UL<<SD_CS_SEL_BIT));
	SD_CS_DIR |= (1UL << SD_CS_BIT);
	DESELECT();

	// enable SSP power
	PCONP |= (1UL << PCONP_PCSPI1);

	// reset pin-functions
	PINSEL1 &= ~(
		(3UL << 2) | // SCK
		(3UL << 4) | // MISO
		(3UL << 6) | // MOSI
		(3UL << 8)   // SSEL
	);
	// set pin-functions to SSP
	PINSEL1 |= (
		(2UL << 2) | // SCK
		(2UL << 4) | // MISO
		(2UL << 6)   // MOSI
		// slave-select handled though GPIO
	);

	// configure and enable as SPI-Master 	
  SSPCR0 = ((8-1)<<SSPCR0_DSS) | (0<<SSPCR0_CPOL) | (0<<SSPCR0_CPHA) | (0<<SSPCR0_SCR);
 	SSPCR1 = (1<<SSPCR1_SSE);

	// slow during initialization
	spiSetSpeed(SPI_SPEED_400kHz);

	// wait for SSP idle)
	while ((SSPSR&(1<<SSPSR_BSY))!=0)
  {
		// wait for busy gone
	}

	while ((SSPSR&(1<<SSPSR_RNE))!=0)
  {
		dummy = SSPDR; // drain SPI RX FIFO
	}
}


static void close_spi( void )
{
	// disable SPI
	SSPCR1 = 0;

	// Pins to GPIO
	PINSEL1 &= ~(
		(3UL << 2) | // SCK
		(3UL << 4) | // MISO
		(3UL << 6) | // MOSI
		(3UL << 8)   // SEL
	);

	// disable SSP power
	PCONP &= ~(1UL << PCONP_PCSPI1);

}

