/**
	@file
	Low-level device driver implementation.
 */

#include "sceptre.h"
#include "device.h"
#include "thermo_device.h"
#include "accelero_device.h"
#include "sdcard_device.h"
#include <reent.h>
#include <errno.h>
#include <string.h>


device_table_array_t device_table_list[DEVICE_MAX_DESCRIPTORS] =
{
  { &thermo_device },
  { &accelero_device },
  //{ &sdcard_device },
  { NULL },
  { NULL }
};


/**
* Search the device table for a device.
*
* @param p_reent      Pointer to a structure that will receive the return value.
* @param device_type  Device type to find.
* @return Index of device on success, -1 if not found.
*/
int device_find(struct _reent *p_reent, uint8_t device_type)
{
  int i;

  // Consider all devices.
  for (i=0; i<DEVICE_MAX_DESCRIPTORS && device_table_list[i].p_item!=0; i++)
  {
    if (device_table_list[i].p_item->device_type==device_type) break;
  }

  // Look to see which device to call.
  if (i>=DEVICE_MAX_DESCRIPTORS || device_table_list[i].p_item==0)
  {
    p_reent->_errno = ENODEV;
    return -1;
  }
  return i;
}


#define MAX_RETRIES   3

/**
* Initialize the device table.
*/
void device_init(void)
{
  int i, retries;
  for (i=0; device_table_list[i].p_item!=0; i++)
  {
    if (device_table_list[i].p_item->init!=NULL)
    {
      for (retries=MAX_RETRIES; retries; retries--)
      {
        // Device will return true if failed to init.
        if (device_table_list[i].p_item->init()==false) break;
      }
      if (retries==0)
      {
        device_table_list[i].p_item->init();
      }
    }
  }
}


/**
* Search the device table for a device by name.
*
* @param p_name  Pointer to the device name to find.
* @param p_len   Length of the device name.
* @return Index of device on success, -1 if not found.
*/
int device_find_by_name(const char *p_name, int *p_len)
{
  int i;

  // All devices must start with a name bracketed with '/'.
  if (p_name[0]=='/')
  {
    // Search the table list.
    for (i=0; device_table_list[i].p_item; i++)
    {
      *p_len = strlen(device_table_list[i].p_item->p_name);
      if (strncmp(device_table_list[i].p_item->p_name,&p_name[1],*p_len)==0)
      {
        // Ensure trailing '/'.
        if (p_name[*p_len+1]=='/')
        {
          // Found it, return the index.
          return i;
        }
      }
    }
  }

  return -1;
}

