/*****************************************************************************
* irq.c: Interrupt handler C file for Philips LPC214x Family Microprocessors
*
* Copyright(C) 2006, Philips Semiconductor
* All rights reserved.
*
* History
* 2005.10.01  ver 1.00    Preliminary version, first Release
*
******************************************************************************/

/**
	@file
	Interrupt handler management.
 */

#include "sceptre.h"


/**
* Default VIC interrupt handler.
*
* This handler is set to deal with spurious interrupts.
* If the IRQ service routine reads the VIC address register
* and no IRQ slot responses as described above, this address
* is returned.
*/
static void irq_vic_default_handler(void) __irq
{
  // If the IRQ is not installed into the VIC, and interrupt occurs, the
  // default interrupt VIC address will be used. This could happen in a race
  // condition. For debugging, use this endless loop to trace back.
  // For more details, see Philips appnote AN10414
  VICVectAddr = 0;  // Acknowledge interrupt.
  while (1);  // Loop forever.
}


/**
* Initialize VIC interrupt controller.
*/
void irq_vic_init(void)
{
  uint32_t i = 0;
  uint32_t *p_vect_addr;
  uint32_t *p_vect_cntl;

  // Initialize VIC.
  VICIntEnClr = 0xffffffff;
  VICVectAddr = 0;
  VICIntSelect = 0;

  // Set all the vector and vector control register to 0.
  for (i=0; i<VIC_SIZE; i++)
  {
    p_vect_addr = (uint32_t *)(VIC_BASE_ADDR + VECT_ADDR_INDEX + i*4);
    p_vect_cntl = (uint32_t *)(VIC_BASE_ADDR + VECT_CNTL_INDEX + i*4);
    *p_vect_addr = 0;
    *p_vect_cntl = 0;
  }

  // Install the default VIC handler here.
  VICDefVectAddr = (uint32_t)irq_vic_default_handler;
}


/**
* Install interrupt handler.
*
* The maximum VIC size is 16 but there are 32 interrupt
* request inputs. Not all of them can be installed into the
* VIC table at the same time. The order of the interrupt 
* request installation is first come, first served.
*
* @param number     Number of interrupt.
* @param p_handler  Address of interrupt handler.
* @return true on success, false if the table was full.
*/
bool_t irq_install(uint32_t number, void *p_handler)
{
  uint32_t i;
  uint32_t *p_vect_addr;
  uint32_t *p_vect_cntl;

  VICIntEnClr = 1<<number; // Disable interrupt.

  for (i=0; i<VIC_SIZE; i++)
  {
    // Find first empty slot in VIC table.
    p_vect_addr = (uint32_t *)(VIC_BASE_ADDR + VECT_ADDR_INDEX + i*4);
    p_vect_cntl = (uint32_t *)(VIC_BASE_ADDR + VECT_CNTL_INDEX + i*4);
    if (*p_vect_addr==(uint32_t)NULL)
    {
    	// Found, install new handler.
      *p_vect_addr = (uint32_t)p_handler;  // Set interrupt vector.
      *p_vect_cntl = (uint32_t)(IRQ_SLOT_EN | number);
      break;
    }
  }
  
  if (i==VIC_SIZE)
  {
    return false;  // Fatal error, can't find empty vector slot.
  }
  
  VICIntEnable = 1<<number;  // Enable interrupt.
  return true;
}


/**
* Uninstall interrupt handler.
*
* Find the interrupt handler installed in the VIC
* based on the interrupt number, set the location
* back to NULL to uninstall it.
*
* @param number     Number of interrupt.
* @return true on success, false if the interrupt number was not found.
*/
bool_t irq_uninstall(uint32_t number)
{
  uint32_t i;
  uint32_t *p_vect_addr;
  uint32_t *p_vect_cntl;

  VICIntEnClr = 1<<number; // Disable interrupt.

  for (i=0; i<VIC_SIZE; i++)
  {
    // Look for the handler of this interrupt number.
    p_vect_addr = (uint32_t *)(VIC_BASE_ADDR + VECT_ADDR_INDEX + i*4);
    p_vect_cntl = (uint32_t *)(VIC_BASE_ADDR + VECT_CNTL_INDEX + i*4);
    if ((*p_vect_cntl&~IRQ_SLOT_EN)==number)
    {
    	// Found, remove it.
      *p_vect_addr = (uint32_t)NULL;  // Clear the entry in the VIC table.
      *p_vect_cntl &= ~IRQ_SLOT_EN;  // Disable SLOT_EN bit.
      break;
    }
  }
  
  if (i==VIC_SIZE)
  {
    return false;  // Interrupt number not found.
  }
  
  VICIntEnable = 1<<number;  // Enable interrupt.
  return true;
}

