/****************************************************************************
*  ds18b20.c:  driver for Maxim DS18B20 thermometer chip.
*
*  Contains generic 1-wire driver.
*  CRC code from Maxim.
*
*  Author: Clemens Valens
*  License: GNU General Public License
*
*  History
*  2009.11.08  ver 1.00
*
*****************************************************************************/

/**
	@file
	Implementation of DS18B20 1-wire thermometer driver.
 */

#include "sceptre.h"
#include "ds18b20.h"

// Pin definitions.
#define DS18B20_PIN  19
#define DS18B20_OUTPUT  IODIR1 |= (1<<DS18B20_PIN)
#define DS18B20_INPUT  IODIR1 &= ~(1<<DS18B20_PIN)
#define DS18B20_SET  IOSET1 = (1<<DS18B20_PIN)
#define DS18B20_CLR  IOCLR1 = (1<<DS18B20_PIN)
#define DS18B20_READ_BIT  (IOPIN1&(1<<DS18B20_PIN))==0? 0 : 1

// Commands.
#define DS18B20_ROM_SEARCH  (0xf0)
#define DS18B20_ROM_READ  (0x33)
#define DS18B20_ROM_MATCH  (0x55)
#define DS18B20_ROM_SKIP  (0xcc)
#define DS18B20_ROM_SEARCH_ALARM  (0xec)
#define DS18B20_FUNCTION_CONVERT  (0x44)
#define DS18B20_FUNCTION_WRITE_SCRATCHPAD  (0x4e)
#define DS18B20_FUNCTION_READ_SCRATCHPAD  (0xbe)
#define DS18B20_FUNCTION_COPY_SCRATCHPAD  (0x48)
#define DS18B20_FUNCTION_RECALL  (0xb8)
#define DS18B20_FUNCTION_READ_POWER_SUPPLY  (0xb4)

#define ONEWIRE_DEVICE_PRESENT  1
#define ONEWIRE_NO_DEVICE  0


const uint8_t onewire_crc_table[] =
{
  0, 94,188,226, 97, 63,221,131,194,156,126, 32,163,253, 31, 65,
  157,195, 33,127,252,162, 64, 30, 95, 1,227,189, 62, 96,130,220,
  35,125,159,193, 66, 28,254,160,225,191, 93, 3,128,222, 60, 98,
  190,224, 2, 92,223,129, 99, 61,124, 34,192,158, 29, 67,161,255,
  70, 24,250,164, 39,121,155,197,132,218, 56,102,229,187, 89, 7,
  219,133,103, 57,186,228, 6, 88, 25, 71,165,251,120, 38,196,154,
  101, 59,217,135, 4, 90,184,230,167,249, 27, 69,198,152,122, 36,
  248,166, 68, 26,153,199, 37,123, 58,100,134,216, 91, 5,231,185,
  140,210, 48,110,237,179, 81, 15, 78, 16,242,172, 47,113,147,205,
  17, 79,173,243,112, 46,204,146,211,141,111, 49,178,236, 14, 80,
  175,241, 19, 77,206,144,114, 44,109, 51,209,143, 12, 82,176,238,
  50,108,142,208, 83, 13,239,177,240,174, 76, 18,145,207, 45,115,
  202,148,118, 40,171,245, 23, 73, 8, 86,180,234,105, 55,213,139,
  87, 9,235,181, 54,104,138,212,149,203, 41,119,244,170, 72, 22,
  233,183, 85, 11,136,214, 52,106, 43,117,151,201, 74, 20,246,168,
  116, 42,200,150, 21, 75,169,247,182,232, 10, 84,215,137,107, 53
};


static uint8_t onewire_crc8 = 0;

uint8_t onewire_do_crc8(uint8_t value);
void onewire_wait_us(uint32_t us);
uint8_t onewire_read_bit(void);
void onewire_write_bit(uint8_t bit);
uint8_t onewire_reset(void);
uint8_t onewire_read_byte(void);
void onewire_write_byte(uint8_t byte);
bool_t onewire_read_serial(uint8_t *p_serial);


/**
* Calculate the CRC8 of the byte value provided.
*
* See Maxim Application Note 27.
*
* @param value  Value to add to crc value.
* @return Updated crc value.
*/
uint8_t onewire_do_crc8(uint8_t value)
{
  onewire_crc8 = onewire_crc_table[onewire_crc8 ^ value];
  return onewire_crc8;
}


/**
* Wait a number of microseconds.
*
* CCLK = 60 MHz
* This probably has to be tuned...
*
* This compiles to
* .L3:
*   sub r3, r3, #1
*   str r3, [sp, #0]
*   ldr r3, [sp, #0]
*   cmp r3, #0
*   bne .L3
*
* @param us  Number of microseconds to wait.
*/
void onewire_wait_us(uint32_t us)
{
  volatile uint32_t count = 4*us;
  while (count>0) count -= 1;
}


/**
* Read one bit on a 1-wire bus.
*
* @return Bit value read.
*/
uint8_t onewire_read_bit(void)
{
  // Issue read time slot.
  DS18B20_OUTPUT;
  DS18B20_CLR;
  // Wait at least 1 us.
  onewire_wait_us(6); // App note 126 recommends 6 us (standard) or 1 us (overdrive).
  DS18B20_INPUT;
  // Wait no more than 15 us.
  onewire_wait_us(9); // App note 126 recommends 9 us (standard) or 1 us (overdrive).
  uint8_t bit = DS18B20_READ_BIT;
  onewire_wait_us(55); // App note 126 recommends 55 us (standard) or 7 us (overdrive).
  return bit;
}


/**
* Write one bit on a 1-wire bus.
*
* @param bit  Value to write.
*/
void onewire_write_bit(uint8_t bit)
{
  uint32_t delay1;
  uint32_t delay2;

  if (bit==0)
  {
    delay1 = 60; // App note 126 recommends 60 us (standard) or 7.5 us (overdrive).
    delay2 = 10; // App note 126 recommends 10 us (standard) or 2.5 us (overdrive).
  }
  else
  {
    delay1 = 6; // App note 126 recommends 6 us (standard) or 1 us (overdrive).
    delay2 = 64; // App note 126 recommends 64 us (standard) or 7.5 us (overdrive).
  }

  // Issue write time slot.
  DS18B20_OUTPUT;
  DS18B20_CLR;

  // Write bit.
  onewire_wait_us(delay1);
  DS18B20_INPUT; // Release bus.
  onewire_wait_us(delay2);
}


/**
* Send a reset on a 1-wire bus.
*
* @return ONEWIRE_DEVICE_PRESENT or ONEWIRE_NO_DEVICE
*/
uint8_t onewire_reset(void)
{
  onewire_wait_us(0); // App note 126 recommends 0 us (standard) or 2.5 us (overdrive).
  // Issue reset time slot.
  DS18B20_OUTPUT;
  DS18B20_CLR;
  onewire_wait_us(480); // App note 126 recommends 480 us (standard) or 70 us (overdrive).
  DS18B20_INPUT; // Release bus.
  onewire_wait_us(70); // App note 126 recommends 70 us (standard) or 8.5 us (overdrive).
  uint8_t bit = DS18B20_READ_BIT; // 0=device present, 1=no device.
  onewire_wait_us(410); // App note 126 recommends 410 us (standard) or 40 us (overdrive).
  return bit==0? ONEWIRE_DEVICE_PRESENT : ONEWIRE_NO_DEVICE;
}


/**
* Read one byte on a 1-wire bus.
*
* @return Byte read.
*/
uint8_t onewire_read_byte(void)
{
  uint8_t result = 0;
  uint8_t mask = 0x01;
  while (mask!=0)
  {
    if (onewire_read_bit()==1) result |= mask;
    mask <<= 1;
  }
  return result;
}


/**
* Write one byte on a 1-wire bus.
*
* @param byte  Value to write.
*/
void onewire_write_byte(uint8_t byte)
{
  uint8_t mask = 0x01;
  while (mask!=0)
  {
    onewire_write_bit(byte&mask);
    mask <<= 1;
  }
}


/**
* Read the serial number of a 1-wire device.
*
* @param p_serial  Address to write serial to.
* @return true on success, false otherwise.
*/
bool_t onewire_read_serial(uint8_t *p_serial)
{
  bool_t result = false;
  if (onewire_reset()==ONEWIRE_DEVICE_PRESENT)
  {
    onewire_crc8 = 0;
    onewire_write_byte(DS18B20_ROM_READ);
    for (int i=0; i<8; i++)
    {
      p_serial[i] = onewire_read_byte();
      if (i<7) onewire_do_crc8(p_serial[i]); // Last byte _is_ crc value.
    }
    result = p_serial[7]==onewire_crc8? true : false;
  }
  return result;
}


/**
* Initialize the DS18B20.
*
* @return true on success, false otherwise.
*/
bool_t ds18b20_init(void)
{
  return onewire_reset()==ONEWIRE_DEVICE_PRESENT? true : false;
}


/**
* Write the scratchpad of the DS18B20.
*
* @param th      Value for Th register.
* @param tl      Value for Tl register.
* @param config  Value for configuration register.
* @return true on success, false otherwise.
*/
bool_t ds18b20_write_scratchpad(uint8_t th, uint8_t tl, uint8_t config)
{
  if (onewire_reset()==ONEWIRE_DEVICE_PRESENT)
  {
    onewire_write_byte(DS18B20_ROM_SKIP);
    onewire_write_byte(DS18B20_FUNCTION_WRITE_SCRATCHPAD);
    onewire_write_byte(th);
    onewire_write_byte(tl);
    onewire_write_byte(config);
    return true;
  }
  return false;
}


/**
* Read the temperature measured by the DS18B20.
*
* @return Temperature in degrees Celsius on success, -100 otherwise.
*/
float ds18b20_read_temperature(void)
{
  float result = -100;
  if (onewire_reset()==ONEWIRE_DEVICE_PRESENT)
  {
    onewire_write_byte(DS18B20_ROM_SKIP);
    onewire_write_byte(DS18B20_FUNCTION_CONVERT);
    while (DS18B20_READ_BIT==0);
    onewire_reset();
    onewire_write_byte(DS18B20_ROM_SKIP);
    onewire_write_byte(DS18B20_FUNCTION_READ_SCRATCHPAD);
    uint16_t iresult = onewire_read_byte();
    iresult += ((uint16_t)onewire_read_byte()<<8);
    result = iresult/16.0;
    //Dunked in ice water the measured temperature varied. 1 ref1=-2, ref2=0.7
  }
  return result;
}


/**
* Read the temperature from the DS18B20.
*
* This function is used by the device driver.
*
* @param p_dst     Address to write data to.
* @param dst_size  Number of bytes to read.
* @return Number of bytes read on success, -1 otherwise.
*/
int ds18b20_read(void *p_dst, uint32_t dst_size)
{
  float result = ds18b20_read_temperature();
  if (result>-100)
  {
    *(float *)p_dst = result;
    return sizeof(float);
  }
  /*int i;
  if (onewire_reset()==ONEWIRE_DEVICE_PRESENT)
  {
    onewire_write_byte(DS18B20_ROM_SKIP);
    onewire_write_byte(DS18B20_FUNCTION_CONVERT);
    while (DS18B20_READ_BIT==0);
    onewire_reset();
    onewire_write_byte(DS18B20_ROM_SKIP);
    onewire_write_byte(DS18B20_FUNCTION_READ_SCRATCHPAD);
    if (dst_size>9) dst_size = 9; // Only 9 bytes in scratchpad.
    for (i=0; i<dst_size; i++)
    {
      p_dst[i] = onewire_read_byte();
    }
    return i;
  }*/
  return -1;
}


/**
* Write the DS18B20.
*
* This function is used by the device driver.
*
* @param p_src     Pointer to data to write.
* @param src_size  Number of bytes to write.
* @return Number of bytes read on success, -1 otherwise.
*/
int ds18b20_write(const uint8_t *p_src, uint32_t src_size)
{
  if (src_size>=3)
  {
    if (ds18b20_write_scratchpad(p_src[0],p_src[1],p_src[2])==true) return 3*sizeof(uint8_t);
  }
  return -1;
}


/**
* Read the serial number of a DS18B20.
*
* @param p_serial  Address to write serial to (8 bytes expected).
* @return true on success, false otherwise.
*/
bool_t ds18b20_read_serial(uint8_t *p_serial)
{
  return onewire_read_serial(p_serial);
}


