/****************************************************************************
*  thermo_device.c: thermometer device driver.
*
*  This is a simple example for a device driver that allows the use of
*  standard C file functions like fopen, fread & fwrite to control it.
*
*  Author: Clemens Valens
*  License: GNU General Public License
*
*  History
*  2009.11.21  ver 1.00
*
*****************************************************************************/

/**
	@file
	Thermometer device driver.
 */

#include "sceptre.h"
#include "device.h"
#include "ds18b20.h"


#ifdef THERMO_DEBUG
#include <stdio.h>
#endif


int thermo_open(struct _reent *p_reent, const char *p_name, int flags, int mode);
_ssize_t thermo_read(struct _reent *p_reent, int file, void *p_dst, size_t dst_size);
_ssize_t thermo_write(struct _reent *p_reent, int file, const void *p_src, size_t src_size);


/**
* A pointer to this structure has to be registered in the device_table_list.
*/
device_table_entry_t thermo_device =
{
  "thermo",             // Device "file" name (without forward slashes).
  SCEPTRE_THERMOMETER,  // Device ID.
  thermo_open,          // open
  NULL,                 // close
  thermo_read,          // read
  thermo_write,         // write
  NULL,                 // init
  NULL                  // ioctl
};


/**
* Open function for the thermometer device driver.
*
* This function gets called when you do something like this:
*
*   FILE *p_thermometer = fopen("/thermo/","rw");
*
* Note the forward slashes surrounding the "file" name. Everything that
* follows the string "/thermo/" is passed to this function in p_name, i.e.
*
*   FILE *p_thermometer = fopen("/thermo/alarm","rw");
*
* will pass "alarm" in p_name to this function.
*
* @param p_reent  Pointer to a structure that will receive the return value.
* @param p_name   Pointer to the name of the file to open.
* @param flags    Flags.
* @param mode     Mode.
* @return A valid handle on success, -1 otherwise.
*/
int thermo_open(struct _reent *p_reent, const char *p_name, int flags, int mode)
{
#ifdef THERMO_DEBUG
  printf("thermo_open(%s,%d,%d)\n",p_name,flags,mode);
#endif
  // Open the device.
  if (ds18b20_init()==true)
  {
    // Return the handle on success.
    return DEVICE(SCEPTRE_THERMOMETER);
  }
#ifdef THERMO_DEBUG
  printf("thermo_open failed\n");
#endif
  p_reent->_errno = ENODEV; // Set reason of failure.
  return -1;
}


/**
* Read function for the thermometer device driver.
*
* This function gets called when you do something like this after
* obtaining a valid FILE handle with fopen():
*
*   uint8_t data[9];
*   fread(data,sizeof(uint8_t),9,p_thermometer);
*
* @param p_reent   Pointer to a structure that will receive the return value.
* @param file      Handle of the file to read from.
* @param p_dst     Address to write the data to.
* @param dst_size  Number of bytes to read.
* @return The number of bytes read on success, -1 otherwise.
*/
_ssize_t thermo_read(struct _reent *p_reent, int file, void *p_dst, size_t dst_size)
{
  int size = ds18b20_read(p_dst,dst_size);
#ifdef THERMO_DEBUG
  //for (int j=0; j<size; j++) printf("%02x ",((uint8_t*)p_dst)[j]);
#endif
  if (size<0) p_reent->_errno = ENODEV; // Set reason of failure.
  return size;
}


/**
* Write function for the thermometer device driver.
*
* This function gets called when you do something like this after
* obtaining a valid FILE handle with fopen():
*
*   uint32_t data = 0x12345678;
*   fwrite(&data,sizeof(uint32_t),1,p_thermometer);
*
* @param p_reent   Pointer to a structure that will receive the return value.
* @param file      Handle of the file to write to.
* @param p_dst     Pointer to the data to write.
* @param dst_size  Number of bytes to write.
* @return The number of bytes written on success, -1 otherwise.
*/
_ssize_t thermo_write(struct _reent *p_reent, int file, const void *p_src, size_t src_size)
{
  int size = ds18b20_write(p_src,src_size);
  if (size<0) p_reent->_errno = ENODEV;  // Set reason of failure.
  return size;
}
