/*****************************************************************************
 *   rtc.c:  Realtime clock C file for Philips LPC214x Family Microprocessors
 *
 *   Copyright(C) 2006, Philips Semiconductor
 *   All rights reserved.
 *
 *   History
 *   2010.01.24  ver 1.01    CPV - Adapted for Sceptre
 *   2005.10.01  ver 1.00    Preliminary version, first release
 *
*****************************************************************************/

// modified by Martin Thomas

/**
	@file
	Implementation of the real time clock (RTC) driver.
 */

#include "sceptre.h"

extern void rtc_handler(void) __irq;

volatile uint32_t rtc_alarm_active = 0;


/**
* RTC interrupt handler, its execution depends on the alarm setting.
*/
void rtc_handler(void) __irq 
{  
  ILR |= ILR_RTCCIF;          /* clear interrupt flag */
  IENABLE;                    /* handles nested interrupt */

  rtc_alarm_active = 1;

  IDISABLE;
  VICVectAddr = 0;            /* Acknowledge Interrupt */
}


/**
* Initialize the real time clock (RTC).
*/
void rtc_init(void)
{
  rtc_alarm_active = 0;

  AMR = 0;
  CIIR = 0;
  CCR = 0;
  
  // Do we need this when running from an external watch crystal?
  PREINT = PREINT_RTC;
  PREFRAC = PREFRAC_RTC;
  
  CCR = CCR_CLKEN;  // Disable CTC-Reset and enable clock with CLKEN.
  CCR |= CCR_CLKSRC; // Switch to 32 kHz xtal.
  PCONP &= ~(1<<9); // Save power by disconnecting RTC from system clock.
}


/**
* Start the real time clock (RTC).
*/
void rtc_start(void) 
{
  CCR |= CCR_CLKEN;
  ILR = ILR_RTCCIF;
}


/**
* Stop the real time clock (RTC).
*/
void rtc_stop(void)
{   
  CCR &= ~CCR_CLKEN;
} 


/**
* Reset the RTC's tick counter
*/
void rtc_clear_tick_counter(void)
{   
  CCR |= CCR_CTCRST; // Reset CTC
}


/**
* Set the real time clock (RTC).
*
* @param p_time  Pointer to the new clock values.
*/
void rtc_set_time(const rtc_time_t *p_time) 
{
  SEC = p_time->seconds;
  MINUTES = p_time->minutes;
  HOUR = p_time->hours;
  DOM = p_time->day_of_month;
  DOW = p_time->day_of_week;
  DOY = p_time->day_of_year;
  MONTH = p_time->month;
  YEAR = p_time->year;
}


/**
* Set the real time clock (RTC) alarm.
*
* @param p_time  Pointer to the new alarm values.
*/
void rtc_set_alarm(const rtc_time_t *p_time) 
{   
  ALSEC = p_time->seconds;
  ALMIN = p_time->minutes;
  ALHOUR = p_time->hours;
  ALDOM = p_time->day_of_month;
  ALDOW = p_time->day_of_week;
  ALDOY = p_time->day_of_year;
  ALMON = p_time->month;
  ALYEAR = p_time->year;
}


/**
* Get the real time clock (RTC).
*
* @param p_time  Pointer to a structure that will receive the current clock values.
*/
void rtc_get_time(rtc_time_t *p_time) 
{
  p_time->seconds = SEC;
  p_time->minutes = MINUTES;
  p_time->hours = HOUR;
  p_time->day_of_month = DOM;
  p_time->day_of_week = DOW;
  p_time->day_of_year = DOY;
  p_time->month = MONTH;
  p_time->year = YEAR;
}


/**
* Set the alarm mask of the real time clock (RTC).
*
* @param mask  Mask for the alarm.
*/
void rtc_set_alarm_mask(uint32_t mask) 
{
  AMR = mask;
}

/*****************************************************************************
**                            End Of File
******************************************************************************/

