/*****************************************************************************
*  timer.c:  Timer C file for Philips LPC214x Family Microprocessors
*
*  Copyright(C) 2006, Philips Semiconductor
*  All rights reserved.
*
*  History
*  2009.11.06  ver 2.00    mods by CPV
*  2005.10.01  ver 1.00    Preliminary version, first Release
*
******************************************************************************/

/**
	@file
	Implementation of the timer functions.
 */

#include "sceptre.h"


volatile uint32_t timer0_counter = 0;
volatile uint32_t timer1_counter = 0;
static uint8_t timer_10ms_prescaler = 10;
static uint32_t timer_10ms_counter = 0;

timer_10ms_hook_t timer_10ms_hook = NULL;
static void timer_tick_10ms(void);


/**
* 10 ms prescaler for system timer.
* If set, calls timer_10ms_hook every 10 ms.
*/
static void timer_tick_10ms(void)
{
  timer_10ms_prescaler -= 1;
  if (timer_10ms_prescaler==0)
  {
    timer_10ms_prescaler = 10;
    timer_10ms_counter += 1;
    if (timer_10ms_hook!=NULL) timer_10ms_hook();
  }
}


/**
* Timer/Counter 0 interrupt handler.
* Executes each t ms @ 60 MHz CPU clock.
*/
static void timer0_handler(void) __irq
{
  T0IR = 1;  // Clear interrupt flag.
  IENABLE;  // Handles nested interrupt.

#if SYSTEM_TIMER==TIMER_ZERO
  timer_tick_10ms();
#endif

  timer0_counter++;

  IDISABLE;
  VICVectAddr = 0;  // Acknowledge interrupt.
}


/**
* Timer/Counter 1 interrupt handler.
* Executes each t ms @ 60 MHz CPU clock.
*/
static void timer1_handler(void) __irq
{
  T1IR = 1;  // Clear interrupt flag.
  IENABLE;   // Handles nested interrupt.

#if SYSTEM_TIMER==TIMER_ONE
  timer_tick_10ms();
#endif

  timer1_counter++;

  IDISABLE;
  VICVectAddr = 0;  // Acknowledge interrupt.
}


/**
* Enable a timer.
*
* @param timer_num  Number (0,1) of timer to enable.
*/
void timer_enable(uint8_t timer_num)
{
  if (timer_num==TIMER_ZERO)
  {
    T0TCR = 1;
  }
  else
  {
    T1TCR = 1;
  }
}


/**
* Disable a timer.
*
* @param timer_num  Number (0,1) of timer to enable.
*/
void timer_disable(uint8_t timer_num)
{
  if (timer_num==TIMER_ZERO)
  {
    T0TCR = 0;
  }
  else
  {
    T1TCR = 0;
  }
}


/**
* Reset a timer.
*
* @param timer_num  Number (0,1) of timer to enable.
*/
void timer_reset(uint8_t timer_num)
{
  uint32_t regVal;

  if (timer_num==TIMER_ZERO)
  {
    regVal = T0TCR;
    regVal |= 0x02;
    T0TCR = regVal;
  }
  else
  {
    regVal = T1TCR;
    regVal |= 0x02;
    T1TCR = regVal;
  }
  return;
}


/**
* Initialize a timer.
*
* @param timer_num  Number (0,1) of timer to initialize.
* @param interval   Timer interval in clock ticks.
* @return true on success, false if the interrupt handler could not be installed.
*/
bool_t timer_init(uint8_t timer_num, uint32_t interval)
{
  if (timer_num==SYSTEM_TIMER)
  {
    timer_10ms_prescaler = 10;
    timer_10ms_counter = 0;
  }

  if (timer_num==TIMER_ZERO)
  {
    timer0_counter = 0;
    T0MR0 = interval;
    T0MCR = 3;  // Interrupt and Reset on MR0.
    return irq_install(TIMER0_INT,(void *)timer0_handler);
  }
  else
  {
    timer1_counter = 0;
    T1MR0 = interval;
    T1MCR = 3;  // Interrupt and Reset on MR0.
    return irq_install(TIMER1_INT,(void *)timer1_handler);
  }
}


/**
* Get the value of a timer counter.
*
* @param timer_num  Number (0,1) of timer to get count of.
*/
uint32_t timer_get_count(uint8_t timer_num)
{
  if (timer_num==TIMER_ZERO) return timer0_counter;
  else return timer1_counter;
}


/**
* Set the value of a timer counter.
*
* @param timer_num  Number (0,1) of timer to set count of.
* @param val        New value for the counter.
*/
void timer_set_count(uint8_t timer_num, uint32_t val)
{
  if (timer_num==TIMER_ZERO) timer0_counter = val;
  else timer1_counter = val;
}


/**
* Get the value of the system timer counter.
*
* @return The value of the system timer counter.
*/
inline uint32_t timer_get_system_count(void)
{
#if SYSTEM_TIMER==TIMER_ZERO
  return timer0_counter;
#elif SYSTEM_TIMER==TIMER_ONE
  return timer1_counter;
#endif
}


/**
* Set the value of the system timer counter.
*
* @param val  The new value of the system timer counter.
*/
inline void timer_set_system_count(uint32_t val)
{
#if SYSTEM_TIMER==TIMER_ZERO
  timer0_counter = val;
#elif SYSTEM_TIMER==TIMER_ONE
  timer1_counter = val;
#endif
}


/**
* Get the value of the 10 ms tick counter.
*
* @return The value of the 10 ms tick counter.
*/
inline uint32_t timer_get_10ms_counter(void)
{
  return timer_10ms_counter;
}


/**
* Set 10 ms hook for system timer.
*
* @param f  Address of the hook function.
*/
void timer_set_10ms_hook(timer_10ms_hook_t f)
{
  timer_10ms_hook = f;
}


/**
* Wait for a specified number of milliseconds.
*
* @param delay_ms  The time to wait in milliseconds (ms).
*/
void timer_wait_ms(uint32_t delay_ms)
{
  uint32_t last_count = timer_get_system_count();
  while (timer_get_system_count()<last_count+delay_ms);
}
