/*****************************************************************************
*   uart.c:  UART driver for Philips LPC214x family microprocessors
*
*   This is a total rewrite of the Philips/WinARM code.
*
*   History
*   2010.01.08  ver 1.01  CPV, added RX hooks to ISR.
*   2009.11.17  ver 1.00  CPV, initial release.
*
******************************************************************************/

/**
	@file
	Implementation of UART0 & UART1 drivers.
 */

#include "sceptre.h"
#include "uart.h"
#include <string.h>


#define UART_BUFFER_SIZE  0x200  /* Must be a power of 2 */
#define UART_BUFFER_MASK  (UART_BUFFER_SIZE-1)

#define UART0_RX_FIFO_SIZE  16  /* As defined in the datasheet. */
#define UART0_TX_FIFO_SIZE  16  /* As defined in the datasheet. */
#define UART1_RX_FIFO_SIZE  16  /* As defined in the datasheet. */
#define UART1_TX_FIFO_SIZE  16  /* As defined in the datasheet. */

/**
* UART buffer.
*/
typedef struct
{
  volatile uint16_t write_ptr; //!< pointer to first empty slot.
  volatile uint16_t read_ptr; //!< pointer to first character.
  uint8_t data[UART_BUFFER_SIZE]; //!< data buffer.
}
uart_buffer_t;

/**
* UART RX & TX buffer.
*/
typedef struct
{
  uart_buffer_t rx; //!< RX buffer.
  uart_buffer_t tx; //!< TX buffer.
  volatile bool_t tx_fifo_empty; //!< Flag to signal that the TX fifo is empty.
}
uart_rx_tx_buffer_t;


// UART0 data & prototypes.
void uart0_handler(void) __irq;
uart_rx_tx_buffer_t uart0_buffer;
uart_rx_hook_t uart0_rx_hook = NULL;


// UART1 data & prototypes.
void uart1_handler(void) __irq;
uart_rx_tx_buffer_t uart1_buffer;
uart_rx_hook_t uart1_rx_hook = NULL;


// uart_buffer function prototypes.
inline void uart_buffer_flush(uart_buffer_t *p_buffer);
inline uint32_t uart_buffer_has_data(uart_buffer_t *p_buffer);
inline bool_t uart_buffer_is_full(uart_buffer_t *p_buffer);
inline bool_t uart_buffer_getc(uint32_t *p_ch, uart_buffer_t *p_buffer);
inline bool_t uart_buffer_putc(uint32_t ch, uart_buffer_t *p_buffer);

/**
* Flush a buffer.
* 
* @param p_buffer  Pointer to the buffer to flush.
*/
inline void uart_buffer_flush(uart_buffer_t *p_buffer)
{
  p_buffer->write_ptr = p_buffer->read_ptr = 0;
}


/**
* Get the number of characters available in a buffer.
* 
* @param p_buffer  Pointer to the buffer.
* @return The number of characters in the buffer.
*/
inline uint32_t uart_buffer_has_data(uart_buffer_t *p_buffer)
{
  return (p_buffer->write_ptr-p_buffer->read_ptr)&UART_BUFFER_MASK;
}


/**
* Check if a buffer is full or not.
* 
* @param p_buffer  Pointer to the buffer to check.
* @return true if the buffer is full, false if it isn't.
*/
inline bool_t uart_buffer_is_full(uart_buffer_t *p_buffer)
{
  return uart_buffer_has_data(p_buffer)==UART_BUFFER_SIZE;
}


/**
* Get the next character from the buffer.
* 
* @param p_ch      Pointer to variable where the character will be stored.
* @param p_buffer  Pointer to the buffer to get the character from.
* @return true if a character was read, false if not.
*/
inline bool_t uart_buffer_getc(uint32_t *p_ch, uart_buffer_t *p_buffer)
{
  if (p_buffer->write_ptr!=p_buffer->read_ptr)
  {
    *p_ch = p_buffer->data[p_buffer->read_ptr];
    p_buffer->read_ptr = (p_buffer->read_ptr+1)&UART_BUFFER_MASK;
    return true;
  }
  return false;
}


/**
* Put a character in the buffer.
* 
* @param ch        Character to put in the buffer.
* @param p_buffer  Pointer to the buffer to put the character in.
* @return true if the character was written, false if it wasn't.
*/
inline bool_t uart_buffer_putc(uint32_t ch, uart_buffer_t *p_buffer)
{
  if (p_buffer->write_ptr!=((p_buffer->read_ptr-1)&UART_BUFFER_MASK))
  {
    p_buffer->data[p_buffer->write_ptr] = (uint8_t) ch;
    p_buffer->write_ptr = (p_buffer->write_ptr+1)&UART_BUFFER_MASK;
    return true;
  }
  return false;
}


/*void uart_set_baudrate(uint32_t baudrate)
{
  uint32_t Fdiv = Fpclk/(16*baudrate);
  U0DLM = Fdiv / 256;
  U0DLL = Fdiv % 256;
}*/


/**
* UART0 interrupt handler
*/
void uart0_handler(void) __irq
{
  uint8_t iir;
  uint8_t lsr;

  IENABLE;  // Handles nested interrupt -> handled in Startup.S.

  // Read interrupt identification register.
  iir = U0IIR;
  iir >>= 1;  // Skip pending bit in IIR.
  iir &= 0x07;  // Check bit 1~3, interrupt identification.

  switch (iir)
  {
    // Receive Line Status
    case IIR_RLS:
      // This interrupt only occurs on OE, PE & FE errors or a break (BI).
      lsr = U0LSR;  // Clear RLS interrupt.
      break;

    // Receive Data Available.
    case IIR_RDA:
      // Split in two to speed up the normal case when there is no RX hook installed.
      if (uart0_rx_hook!=NULL)
      {
        // Read as much data from the RX fifo as possible, pass through RX hook.
        while ((U0LSR&LSR_RDR)!=0)
        {
          uint32_t ch = U0RBR;
          // Do not put char in buffer if the hook returned true.
          if (uart0_rx_hook(ch)==false) uart_buffer_putc(ch,&uart0_buffer.rx);
        }
      }
      else
      {
        // Read as much data from the RX fifo as possible.
        while ((U0LSR&LSR_RDR)!=0)
        {
          uart_buffer_putc(U0RBR,&uart0_buffer.rx);
        }
      }
      break;

    // Character Time-out Indicator.
    case IIR_CTI:
      // There is still an unread character left in the fifo, go get it.
      uart_buffer_putc(U0RBR,&uart0_buffer.rx);
      break;

    // Transmit Holding Register Empty.
    case IIR_THRE:
    {
      uart0_buffer.tx_fifo_empty = true; // Mark fifo as empty.
      // Fill the tx fifo with as much data as possible.
      uint8_t count = UART0_TX_FIFO_SIZE;
      while (count>0)
      {
        uint32_t ch;
        if (uart_buffer_getc(&ch,&uart0_buffer.tx)==false) break;
        U0THR = ch;  // Put character in fifo.
        uart0_buffer.tx_fifo_empty = false;  // Mark fifo as no longer empty.
        count -= 1;
      }
      break;
    }
  }

  IDISABLE;  // Handled in Startup.S.
  VICVectAddr = 0;  // Acknowledge interrupt.
}


/**
* Initialize UART0, setup pin select, clock, parity, stop bits, FIFO, etc.
*
* @param baudrate   Baudrate
* @param data_bits  Number of data bits [5,6,7,8]
* @param parity     Parity ['n','N,'o','O','e','E','1','0']
* @param stop_bits  Number of stop bits [1,2]
*
* @return   false if the interrupt handler could not be installed in the VIC table,
*           true otherwise.
*/
bool_t uart0_init(uint32_t baudrate, uint8_t data_bits, char parity, uint8_t stop_bits)
{
  uint32_t dummy;

  // Clear RX & TX buffers.
  memset(&uart0_buffer,0,sizeof(uart0_buffer));
  uart0_buffer.tx_fifo_empty = true;

  PINSEL0 = (PINSEL0&0xfffffff0)|0x00000005;  // Enable RxD0 and TxD0

  // disable any interrupts currently enabled.
  U0IER = 0;
  // clear any possible interrupt sources.
  dummy = U0IIR;
  dummy = U0RBR;
  dummy = U0LSR;

  // Setup communication parameters.
  uint32_t p = 0;
  if (parity=='o' || parity=='O') p = 0x08;
  else if (parity=='e' || parity=='E') p = 0x18;
  else if (parity=='1') p = 0x28;
  else if (parity=='0') p = 0x38;
  // Setup data bits [5,8], stop bits [1,2] & parity ['n','o','e','1','0'].
  U0LCR = 0x80 + ((data_bits-5)&0x3) + (((stop_bits-1)&0x1)<<1) + p;
  // Setup baud rate.
//TODO: replace with more accurate calculation.
  uint32_t Fdiv = (Fpclk/16)/baudrate;
  U0DLM = Fdiv / 256;
  U0DLL = Fdiv % 256;

  // Better precision for 115200 @ PCLK=60 MHz
	//U0DLM = 0;
	//U0DLL = 30;
	//U0FDR = (12<<4) | 1;

  U0LCR &= 0x7f;  // DLAB = 0.

  // Enable and reset TX and RX FIFO, RX fifo trigger level is 1 character.
  U0FCR = 0x07;

  if (irq_install(UART0_INT,(void*)uart0_handler)==false)
  {
    return false;
  }

  // Enable RX, TX & line status interrupts.
  U0IER = IER_RBR | IER_THRE | IER_RLS;
  return true;
}


/**
* Send a block of data using UART0.
*
* @param p_data      Pointer to data.
* @param data_size   Number of bytes to send.
* @param cr_to_crlf  Convert CR to CRLF or not.
*
* @return  The number of characters sent.
*
*/
uint32_t uart0_send(const uint8_t *p_data, uint32_t data_size, bool_t cr_to_crlf)
{
  uint32_t i = 0;
  while (i<data_size)
  {
    if (p_data[i]=='\n' && cr_to_crlf==true)
    {
      // Translate "\n" to "\r\n"
      if (uart0_putc('\r')==false) break;
    }
    if (uart0_putc(p_data[i])==false) break;
    i += 1;
  }
  return i;
}


/**
* Send one character using UART0.
*
* @param ch  Character to send.
*
* @return true on success, false otherwise.
*/
bool_t uart0_putc(uint32_t ch)
{
  if (uart0_buffer.tx_fifo_empty==true)
  {
    U0THR = ch;  // Send character right away.
    uart0_buffer.tx_fifo_empty = false;  // Mark fifo as no longer empty.
    return true;
  }
  else return uart_buffer_putc(ch,&uart0_buffer.tx);  // Put character in tx buffer.
}


/**
* Get a character from UART0.
*
* @param p_ch           Pointer to variable where the received character will be stored.
* @param timeout_in_ms  Time (ms) to wait for a character to come in, 0 equals no wait.
*
* @return true on success, false otherwise.
*/
bool_t uart0_getc(uint32_t *p_ch, uint32_t timeout_in_ms)
{
  // Wait for a character to come in, but do not wait forever.
  uint32_t last_count = timer_get_system_count();
  while (uart_buffer_has_data(&uart0_buffer.rx)==0 &&
         timer_get_system_count()-last_count<timeout_in_ms);
  return uart_buffer_getc(p_ch,&uart0_buffer.rx); // Get character (if any).
}


/**
* Clear the TX & RX data buffers of UART0.
*/
void uart0_flush_buffers(void)
{
  uart_buffer_flush(&uart0_buffer.rx);
  uart_buffer_flush(&uart0_buffer.tx);
}


/**
* Set RX hook for UART1.
*
* @param f  Pointer to hook function.
*/
void uart0_set_rx_hook(uart_rx_hook_t f)
{
  uart0_rx_hook = f;
}


///////////////////////////////////
///////////// UART1 ///////////////
///////////////////////////////////


/**
* UART0 interrupt handler
*/
void uart1_handler(void) __irq
{
  uint8_t iir;
  uint8_t lsr;

  IENABLE;  // Handles nested interrupt -> handled in Startup.S.

  // Read interrupt identification register.
  iir = U1IIR;
  iir >>= 1;  // Skip pending bit in IIR.
  iir &= 0x07;  // Check bit 1~3, interrupt identification.

  switch (iir)
  {
    // Receive Line Status
    case IIR_RLS:
      // This interrupt only occurs on OE, PE & FE errors or a break (BI).
      lsr = U1LSR;  // Clear RLS interrupt.
      break;

    // Receive Data Available.
    case IIR_RDA:
      // Split in two to speed up the normal case when there is no RX hook installed.
      if (uart1_rx_hook!=NULL)
      {
        // Read as much data from the RX fifo as possible, pass through RX hook.
        while ((U1LSR&LSR_RDR)!=0)
        {
          uint32_t ch = U1RBR;
          // Do not put char in buffer if the hook returned true.
          if (uart1_rx_hook(ch)==false) uart_buffer_putc(ch,&uart1_buffer.rx);
        }
      }
      else
      {
        // Read as much data from the RX fifo as possible.
        while ((U1LSR&LSR_RDR)!=0)
        {
          uart_buffer_putc(U1RBR,&uart1_buffer.rx);
        }
      }
      break;

    // Character Time-out Indicator.
    case IIR_CTI:
      // There is still an unread character left in the fifo, go get it.
      uart_buffer_putc(U1RBR,&uart1_buffer.rx);
      break;

    // Transmit Holding Register Empty.
    case IIR_THRE:
    {
      uart1_buffer.tx_fifo_empty = true; // Mark fifo as empty.
      // Fill the tx fifo with as much data as possible.
      uint8_t count = UART1_TX_FIFO_SIZE;
      while (count>0)
      {
        uint32_t ch;
        if (uart_buffer_getc(&ch,&uart1_buffer.tx)==false) break;
        U1THR = ch;  // Put character in fifo.
        uart1_buffer.tx_fifo_empty = false;  // Mark fifo as no longer empty.
        count -= 1;
      }
      break;
    }
  }

  IDISABLE;  // Handled in Startup.S.
  VICVectAddr = 0;  // Acknowledge interrupt.
}


/**
* Initialize UART1, setup pin select, clock, parity, stop bits, FIFO, etc.
*
* @param baudrate   Baudrate
* @param data_bits  Number of data bits [5,6,7,8]
* @param parity     Parity ['n','N,'o','O','e','E','1','0']
* @param stop_bits  Number of stop bits [1,2]
*
* @return   false if the interrupt handler could not be installed in the VIC table,
*           true otherwise.
*/
bool_t uart1_init(uint32_t baudrate, uint8_t data_bits, char parity, uint8_t stop_bits)
{
  uint32_t dummy;

  // Clear RX & TX buffers.
  memset(&uart1_buffer,0,sizeof(uart1_buffer));
  uart1_buffer.tx_fifo_empty = true;

  PINSEL0 = (PINSEL0&0xfff0ffff)|0x00050000;  // Enable RxD1 and TxD1

  // disable any interrupts currently enabled.
  U1IER = 0;
  // clear any possible interrupt sources.
  dummy = U1IIR;
  dummy = U1RBR;
  dummy = U1LSR;

  // Setup communication parameters.
  uint32_t p = 0;
  if (parity=='o' || parity=='O') p = 0x08;
  else if (parity=='e' || parity=='E') p = 0x18;
  else if (parity=='1') p = 0x28;
  else if (parity=='0') p = 0x38;
  // Setup data bits [5,8], stop bits [1,2] & parity ['n','o','e','1','0'].
  U1LCR = 0x80 + ((data_bits-5)&0x3) + (((stop_bits-1)&0x1)<<1) + p;
  // Setup baud rate.
  uint32_t Fdiv = (Fpclk/16)/baudrate;
  U1DLM = Fdiv / 256;
  U1DLL = Fdiv % 256;
  U1LCR &= 0x7f;  // DLAB = 0.

  // Enable and reset TX and RX FIFO, RX fifo trigger level is 1 character.
  U1FCR = 0x07;

  if (irq_install(UART1_INT,(void*)uart1_handler)==false)
  {
    return false;
  }

  // Enable RX, TX & line status interrupts.
  U1IER = IER_RBR | IER_THRE | IER_RLS;
  return true;
}


/**
* Send a block of data using UART1.
*
* @param p_data      Pointer to data.
* @param data_size   Number of bytes to send.
* @param cr_to_crlf  Convert CR to CRLF or not.
*
* @return  The number of characters sent.
*/
uint32_t uart1_send(const uint8_t *p_data, uint32_t data_size, bool_t cr_to_crlf)
{
  uint32_t i = 0;
  while (i<data_size)
  {
    if (p_data[i]=='\n' && cr_to_crlf==true)
    {
      // Translate "\n" to "\r\n"
      if (uart1_putc('\r',false)==false) break;
    }
    if (uart1_putc(p_data[i],false)==false) break;
    i += 1;
  }
  return i;
}


/**
* Send one character using UART1.
*
* @param ch         Character to send.
* @param right_now  If true will send character immediately.
*
* @return true on success, false otherwise.
*/
bool_t uart1_putc(uint32_t ch, bool_t right_now)
{
  if (right_now==true)
  {
    U1THR = ch;  // Send character right away.
  }
  else
  {
    if (uart1_buffer.tx_fifo_empty==true)
    {
      U1THR = ch;  // Send character right away.
      uart1_buffer.tx_fifo_empty = false;  // Mark fifo as no longer empty.
    }
    else return uart_buffer_putc(ch,&uart1_buffer.tx);  // Put character in tx buffer.
  }
  return true;
}


/**
* Get a character from UART1.
*
* @param p_ch           Pointer to variable where the received character will be stored.
* @param timeout_in_ms  Time (ms) to wait for a character to come in, 0 equals no wait.
*
* @return true on success, false otherwise.
*/
bool_t uart1_getc(uint32_t *p_ch, uint32_t timeout_in_ms)
{
  // Wait for a character to come in, but do not wait forever.
  uint32_t last_count = timer_get_system_count();
  while (uart_buffer_has_data(&uart1_buffer.rx)==0 &&
         timer_get_system_count()-last_count<timeout_in_ms);
  return uart_buffer_getc(p_ch,&uart1_buffer.rx); // Get character (if any).
}


/**
* Clear the TX & RX data buffers of UART1.
*/
void uart1_flush_buffers(void)
{
  uart_buffer_flush(&uart1_buffer.rx);
  uart_buffer_flush(&uart1_buffer.tx);
}


/**
* Set RX hook for UART1.
*
* @param f  Pointer to hook function.
*/
void uart1_set_rx_hook(uart_rx_hook_t f)
{
  uart1_rx_hook = f;
}
